#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from math import isnan, nan
from typing import cast

import numpy as np
from ax.adapter import Adapter
from ax.adapter.base import DataLoaderConfig
from ax.adapter.cross_validation import cross_validate
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.registry import Generators, MBM_X_trans, Y_trans
from ax.adapter.torch import TorchAdapter
from ax.adapter.transforms.map_key_to_float import MapKeyToFloat
from ax.api.client import Client
from ax.api.configs import RangeParameterConfig
from ax.api.utils.generation_strategy_dispatch import _get_sobol_node
from ax.core.map_data import MAP_KEY
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    observations_from_data,
)
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.trial import Trial
from ax.core.trial_status import TrialStatus
from ax.early_stopping.strategies import PercentileEarlyStoppingStrategy
from ax.generation_strategy.center_generation_node import CenterGenerationNode
from ax.generation_strategy.generation_node import GenerationNode
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generators.base import Generator
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.surrogate import ModelConfig, SurrogateSpec
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment_with_timestamp_map_metric
from botorch.acquisition.logei import qLogExpectedImprovement
from botorch.models.gp_regression import SingleTaskGP
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance, none_throws


class ClientTest(TestCase):
    def setUp(self) -> None:
        super().setUp()

        seed = 8888
        range_parameters = [
            RangeParameterConfig(name="width", parameter_type="float", bounds=(1, 20)),
            RangeParameterConfig(name="height", parameter_type="float", bounds=(1, 20)),
        ]

        self.initialization_budget = 10
        self.max_steps = 15
        self.metric_name = "test_loss"
        self.num_parameters = len(range_parameters)

        objective = f"-{self.metric_name}"

        surrogate_spec = SurrogateSpec(
            model_configs=[ModelConfig(botorch_model_class=SingleTaskGP)]
        )
        generator_spec = GeneratorSpec(
            generator_enum=Generators.BOTORCH_MODULAR,
            model_kwargs={
                "surrogate_spec": surrogate_spec,
                "botorch_acqf_class": qLogExpectedImprovement,
                "transforms": MBM_X_trans + Y_trans,
                "data_loader_config": DataLoaderConfig(
                    fit_only_completed_map_metrics=False,
                    latest_rows_per_group=1,
                ),
            },
        )

        generation_strategy = self._construct_generation_strategy(
            generator_spec=generator_spec,
            initialization_budget=self.initialization_budget,
            seed=seed,
        )
        early_stopping_strategy = PercentileEarlyStoppingStrategy(
            min_progression=3,
            min_curves=3,
        )

        self.client = Client()
        self.client.configure_experiment(parameters=range_parameters)
        self.client.configure_optimization(objective=objective)
        self.client.set_generation_strategy(generation_strategy=generation_strategy)
        self.client.set_early_stopping_strategy(
            early_stopping_strategy=early_stopping_strategy
        )

    @staticmethod
    def _construct_generation_strategy(
        generator_spec: GeneratorSpec,
        initialization_budget: int,
        seed: int,
    ) -> GenerationStrategy:
        """Constructs a Center + Sobol + Modular BoTorch `GenerationStrategy`
        using the provided `generator_spec` for the Modular BoTorch node.
        """
        sobol_node = _get_sobol_node(
            initialization_budget=initialization_budget,
            min_observed_initialization_trials=None,
            initialize_with_center=True,
            use_existing_trials_for_initialization=True,
            allow_exceeding_initialization_budget=False,
            initialization_random_seed=seed,
        )
        center_node = CenterGenerationNode(next_node_name=sobol_node.node_name)
        botorch_node = GenerationNode(
            node_name="MBM",
            generator_specs=[generator_spec],
            should_deduplicate=True,
        )
        return GenerationStrategy(
            name=f"Center+Sobol+{botorch_node.node_name}",
            nodes=[center_node, sobol_node, botorch_node],
        )

    @staticmethod
    def _loss_fn(step: int, width: float, height: float) -> float:
        return 100.0 / (10.0 + width * step) + 0.1 * height

    def _simulate(
        self,
        with_early_stopping: bool = True,
        attach_with_progression: bool = True,
        complete_with_progression: bool = True,
    ) -> None:
        """Simulate typical usage of Client API."""
        for _ in range(self.initialization_budget):
            (trial_data,) = self.client.get_next_trials(max_trials=1).items()
            trial_index, parameters = trial_data
            kwargs: dict[str, float] = {k: float(v) for k, v in parameters.items()}

            stopped: bool = False
            if with_early_stopping:
                for i in range(self.max_steps - 1):
                    step = i + 1
                    result = self._loss_fn(step=step, **kwargs)
                    self.client.attach_data(
                        trial_index=trial_index,
                        raw_data={self.metric_name: result},
                        progression=step if attach_with_progression else None,
                    )
                    if stopped := self.client.should_stop_trial_early(
                        trial_index=trial_index
                    ):
                        self.client.mark_trial_early_stopped(trial_index=trial_index)
                        break

            if not stopped:
                result = self._loss_fn(step=self.max_steps, **kwargs)
                self.client.complete_trial(
                    trial_index=trial_index,
                    raw_data={self.metric_name: result},
                    progression=self.max_steps if complete_with_progression else None,
                )

    def _test_no_early_stopping(self, with_progression: bool) -> None:
        self._simulate(
            with_early_stopping=False,
            attach_with_progression=with_progression,
            complete_with_progression=with_progression,
        )

        # ensure there are no early-stopped trials for the purposes of this test
        self.assertEqual(
            len(self.client._experiment.trials_by_status[TrialStatus.EARLY_STOPPED]),
            0,
        )

        self.client.get_next_trials(max_trials=1)

        adapter = assert_is_instance(
            self.client._generation_strategy.adapter, TorchAdapter
        )
        generator = assert_is_instance(adapter.generator, BoTorchGenerator)
        surrogate = generator.surrogate
        (dataset,) = surrogate.training_data

        # the transform behaves as a no-op (list of parameters to add to
        # search space is empty)
        self.assertListEqual(
            assert_is_instance(
                adapter.transforms["MapKeyToFloat"], MapKeyToFloat
            )._parameter_list,
            [],
        )

        # progression information is omitted from data propagated to the model
        self.assertListEqual(dataset.feature_names, ["width", "height"])

        # Check that cross validation works.
        cross_validate(model=adapter)

    def _test_early_stopping(self, complete_with_progression: bool) -> None:
        self._simulate(
            with_early_stopping=True,
            attach_with_progression=True,
            complete_with_progression=complete_with_progression,
        )

        # ensure there are early-stopped trials for the purposes of this test
        self.assertGreater(
            len(self.client._experiment.trials_by_status[TrialStatus.EARLY_STOPPED]),
            0,
        )

        (trial_index,) = self.client.get_next_trials(max_trials=1)

        adapter = assert_is_instance(
            self.client._generation_strategy.adapter, TorchAdapter
        )
        generator = assert_is_instance(adapter.generator, BoTorchGenerator)
        surrogate = generator.surrogate
        (dataset,) = surrogate.training_data

        # check that the data being fed includes all trials, including
        # early-stopped trials
        self.assertEqual(
            dataset.X.shape,
            (self.initialization_budget, self.num_parameters + 1),
        )

        # check that the data being fed to the model is properly
        # contextualized with progression information
        self.assertListEqual(dataset.feature_names, ["width", "height", "step"])

        trial = cast(Trial, self.client._experiment.trials[trial_index])
        generator_run = none_throws(trial.generator_run)
        candidate_metadata_by_arm_signature = none_throws(
            generator_run.candidate_metadata_by_arm_signature
        )
        signature = none_throws(trial.arm).signature
        candidate_metadata = none_throws(candidate_metadata_by_arm_signature[signature])

        # check that candidate is generated at the target progression
        self.assertEqual(int(candidate_metadata["step"]), self.max_steps)

        # Check that cross validation works.
        cross_validate(model=adapter)

    def test_no_early_stopping_with_progression(self) -> None:
        self._test_no_early_stopping(with_progression=True)

    def test_no_early_stopping_no_progression(self) -> None:
        self._test_no_early_stopping(with_progression=False)

    def test_early_stopping_with_final_progression(self) -> None:
        self._test_early_stopping(complete_with_progression=True)


class MapKeyToFloatTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_branin_experiment_with_timestamp_map_metric(
            with_trials_and_data=True,
        )
        self.search_space = self.experiment.search_space
        self.adapter = Adapter(experiment=self.experiment, generator=Generator())
        self.observations = observations_from_data(
            experiment=self.experiment,
            data=self.experiment.fetch_data(),
            latest_rows_per_group=None,
        )
        self.map_key: str = MAP_KEY
        # Does not require explicitly specifying `config`.
        self.t = MapKeyToFloat(observations=self.observations, adapter=self.adapter)

    def test_Init(self) -> None:
        # Check for error if adapter & parameters are not provided.
        with self.assertWarnsRegex(Warning, "optimization config"):
            MapKeyToFloat(observations=self.observations)

        experiment_data = extract_experiment_data(
            experiment=self.experiment,
            data_loader_config=DataLoaderConfig(fit_only_completed_map_metrics=False),
        )
        t2 = MapKeyToFloat(experiment_data=experiment_data, adapter=self.adapter)
        for t in (self.t, t2):
            # Check for default initialization
            self.assertEqual(len(t._parameter_list), 1)
            (p,) = t._parameter_list
            self.assertEqual(p.name, self.map_key)
            self.assertEqual(p.parameter_type, ParameterType.FLOAT)
            self.assertEqual(p.lower, 0.0)
            self.assertEqual(p.upper, 4.0)
            self.assertFalse(p.log_scale)

        # specifying a parameter name that is not in the observation features' metadata
        with self.assertRaisesRegex(KeyError, "'baz'"):
            MapKeyToFloat(
                observations=self.observations,
                config={"parameters": {"baz": {}}},
            )

        # test that one is able to override default config
        with self.subTest(msg="override default config"):
            t = MapKeyToFloat(
                observations=self.observations,
                config={"parameters": {self.map_key: {"lower": 0.1}}},
            )
            self.assertDictEqual(t.parameters, {self.map_key: {"lower": 0.1}})
            self.assertEqual(len(t._parameter_list), 1)

            p = t._parameter_list[0]

            self.assertEqual(p.name, self.map_key)
            self.assertEqual(p.parameter_type, ParameterType.FLOAT)
            self.assertEqual(p.lower, 0.1)
            self.assertEqual(p.upper, 4.0)
            self.assertFalse(p.log_scale)

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)

        self.assertSetEqual(set(ss2.parameters), {"x1", "x2", self.map_key})

        p = assert_is_instance(ss2.parameters[self.map_key], RangeParameter)
        self.assertEqual(p.name, self.map_key)
        self.assertEqual(p.parameter_type, ParameterType.FLOAT)
        self.assertEqual(p.lower, 0.0)
        self.assertEqual(p.upper, 4.0)
        self.assertFalse(p.log_scale)

    def test_TransformObservationFeatures(self) -> None:
        # NaN progressions get filled in with the upper bound,
        # and then upon untransforming, the value remains non-NaN and won't
        # match its original value.
        with self.subTest("Non-NaN observation features"):
            keep_indices = [
                i
                for i, obs in enumerate(self.observations)
                if not isnan(obs.features.metadata["step"])
            ]
            observation_features = [self.observations[i].features for i in keep_indices]
            obs_ft2 = deepcopy(observation_features)
            obs_ft2 = self.t.transform_observation_features(obs_ft2)

            expected = []
            for i in keep_indices:
                obs = self.observations[i]
                obsf = obs.features.clone()
                obsf.parameters[self.map_key] = obsf.metadata.pop(self.map_key)
                expected.append(obsf)

            self.assertEqual(obs_ft2, expected)
            obs_ft2 = self.t.untransform_observation_features(obs_ft2)
            self.assertEqual(obs_ft2, observation_features)

        with self.subTest("NaN observation features"):
            keep_indices = [
                i
                for i, obs in enumerate(self.observations)
                if isnan(obs.features.metadata["step"])
            ]
            observation_features = [self.observations[i].features for i in keep_indices]
            obs_ft2 = deepcopy(observation_features)
            obs_ft2 = self.t.transform_observation_features(obs_ft2)

            # upper bound
            expected = []
            for i in keep_indices:
                obs = self.observations[i]
                obsf = obs.features.clone()
                obsf.parameters[self.map_key] = 4.0
                obsf.metadata = {}
                expected.append(obsf)

            self.assertEqual(obs_ft2, expected)
            untransformed = self.t.untransform_observation_features(obs_ft2)
            expected = observation_features
            for obs in expected:
                obs.metadata["step"] = 4

            self.assertEqual(untransformed, observation_features)

    def test_mixed_nan_progression(self) -> None:
        observations = [
            Observation(
                data=ObservationData(
                    metric_names=["metric1", "metric2"],
                    means=np.array([1.0, 2.0]),
                    covariance=np.diag([1.0, 2.0]),
                ),
                features=ObservationFeatures(
                    parameters={"x1": 5.0, "x2": 2.0},
                    metadata={self.map_key: 10.0},
                ),
            ),
            Observation(
                data=ObservationData(
                    metric_names=["metric3"],
                    means=np.array([30.0]),
                    covariance=np.array([[30.0]]),
                ),
                features=ObservationFeatures(
                    parameters={"x1": 5.0, "x2": 2.0},
                    metadata={self.map_key: np.nan},
                ),
            ),
            Observation(
                data=ObservationData(
                    metric_names=["x1", "x2"],
                    means=np.array([1.5, 2.5]),
                    covariance=np.diag([1.0, 2.0]),
                ),
                features=ObservationFeatures(
                    parameters={"x1": 5.0, "x2": 2.0},
                    metadata={self.map_key: 20.0},
                ),
            ),
        ]
        observation_features = [obs.features for obs in observations]

        t = MapKeyToFloat(observations=observations, adapter=self.adapter)

        self.assertEqual(len(t._parameter_list), 1)
        (p,) = t._parameter_list
        self.assertEqual(p.name, self.map_key)
        self.assertEqual(p.parameter_type, ParameterType.FLOAT)
        self.assertEqual(p.lower, 10.0)
        self.assertEqual(p.upper, 20.0)

        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = t.transform_observation_features(obs_ft2)

        self.assertEqual(
            obs_ft2,
            [
                ObservationFeatures(
                    parameters={"x1": 5.0, "x2": 2.0, self.map_key: 10.0},
                    metadata={},
                ),
                ObservationFeatures(
                    parameters={"x1": 5.0, "x2": 2.0, self.map_key: 20.0},
                    metadata={},
                ),
                ObservationFeatures(
                    parameters={"x1": 5.0, "x2": 2.0, self.map_key: 20.0},
                    metadata={},
                ),
            ],
        )

    def test_TransformObservationFeaturesKeyNotInMetadata(self) -> None:
        observation_features = [obs.features for obs in self.observations]
        obs_ft2 = deepcopy(observation_features)
        # remove the key from metadata dicts
        for obsf in obs_ft2:
            obsf.metadata.pop(self.map_key)
            # To avoid this being treated as empty metadata.
            # In typical experiment, trial completion timestamp would be here.
            obsf.metadata["dummy"] = 1.0
        # should be exactly one parameter
        (p,) = self.t._parameter_list
        # Transform fills missing values with the upper bound.
        tf_obs_ft = self.t.transform_observation_features(obs_ft2)
        for obs in tf_obs_ft:
            self.assertEqual(obs.parameters[p.name], p.upper)
            self.assertEqual(obs.metadata, {"dummy": 1.0})

    def test_constant_progression(self) -> None:
        for constant in (23, np.nan):
            with self.subTest(msg=f"{constant=}"):
                observations = deepcopy(self.observations)
                for obs in observations:
                    obs.features.metadata[self.map_key] = constant
                observation_features = [obs.features for obs in observations]
                t = MapKeyToFloat(observations=observations, adapter=self.adapter)

                # forward and reverse transforms are identity/no-ops
                obs_ft2 = deepcopy(observation_features)
                obs_ft2 = t.transform_observation_features(obs_ft2)
                self.assertEqual(obs_ft2, observation_features)
                obs_ft2 = t.untransform_observation_features(obs_ft2)
                self.assertEqual(obs_ft2, observation_features)

    def test_TransformObservationFeaturesWithEmptyMetadata(self) -> None:
        # undefined metadata
        obsf = ObservationFeatures(
            trial_index=42,
            parameters={"x1": 1.0, "x2": 2.0},
            metadata=None,
        )
        self.t.transform_observation_features([obsf])
        self.assertEqual(
            obsf,
            ObservationFeatures(
                trial_index=42,
                parameters={"x1": 1.0, "x2": 2.0, self.map_key: 4.0},
                metadata={},
            ),
        )
        # empty metadata
        obsf = ObservationFeatures(
            trial_index=42,
            parameters={"x1": 1.0, "x2": 2.0},
            metadata={},
        )
        self.t.transform_observation_features([obsf])
        self.assertEqual(
            obsf,
            ObservationFeatures(
                trial_index=42,
                parameters={"x1": 1.0, "x2": 2.0, self.map_key: 4.0},
                metadata={},
            ),
        )

    def test_TransformObservationFeaturesWithEmptyParameters(self) -> None:
        obsf = ObservationFeatures(parameters={})
        self.t.transform_observation_features([obsf])

        p = self.t._parameter_list[0]
        self.assertEqual(
            obsf,
            ObservationFeatures(parameters={self.map_key: p.upper}),
        )

    def test_with_different_map_key(self) -> None:
        observations = [
            Observation(
                features=ObservationFeatures(
                    trial_index=0,
                    parameters={"x1": width, "x2": height},
                    metadata={"map_key": timestamp},
                ),
                data=ObservationData(
                    metric_names=[], means=np.array([]), covariance=np.empty((0, 0))
                ),
            )
            for width, height, timestamp in (
                (0.0, 1.0, 12345.0),
                (0.1, 0.9, 12346.0),
            )
        ]
        t = MapKeyToFloat(
            observations=observations,
            config={"parameters": {"map_key": {"log_scale": False}}},
        )
        self.assertEqual(t.parameters, {"map_key": {"log_scale": False}})
        self.assertEqual(len(t._parameter_list), 1)
        tf_obs_ft = t.transform_observation_features(
            [obs.features for obs in observations]
        )
        self.assertEqual(
            tf_obs_ft[0].parameters, {"x1": 0.0, "x2": 1.0, "map_key": 12345.0}
        )
        self.assertEqual(
            tf_obs_ft[1].parameters, {"x1": 0.1, "x2": 0.9, "map_key": 12346.0}
        )

    def test_transform_experiment_data(self) -> None:
        # First, set up a case with no NaNs.
        data = self.experiment.lookup_data()
        data.map_df["step"] = data.map_df["step"].fillna(0)
        experiment_data = extract_experiment_data(
            experiment=self.experiment,
            data=data,
            data_loader_config=DataLoaderConfig(fit_only_completed_map_metrics=False),
        )
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(experiment_data)
        )
        # Check that arm_data is not modified.
        assert_frame_equal(experiment_data.arm_data, transformed_data.arm_data)

        # Since there are no NaNs, observation data is also not modified.
        assert_frame_equal(
            experiment_data.observation_data, transformed_data.observation_data
        )
        # Test with NaNs.
        experiment_data = extract_experiment_data(
            experiment=self.experiment,
            data_loader_config=DataLoaderConfig(fit_only_completed_map_metrics=False),
        )
        # The timestamp is always NaN for 'branin'
        actual = experiment_data.observation_data.index.get_level_values(
            "step"
        ).to_numpy()
        expected_timestamp = np.array([nan, 4.0, nan, 2.0, nan, 0.0])
        self.assertTrue(np.array_equal(actual, expected_timestamp, equal_nan=True))
        # Transform and check that the index is filled with the upper bound.
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(experiment_data)
        )
        transformed_timestamp = (
            transformed_data.observation_data.index.get_level_values("step").tolist()
        )
        upper = self.t._parameter_list[0].upper
        expected_trans_timestamp = [upper, 4.0, upper, 2.0, upper, 0.0]
        self.assertEqual(transformed_timestamp, expected_trans_timestamp)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.benchmark.problems.surrogate.hss.mnist_surrogate import (
    get_mnist_surrogate_benchmark,
)
from ax.utils.common.testutils import TestCase


class TestMNISTSurrogate(TestCase):
    def test_mnist_surrogate(self) -> None:
        cases = [
            (
                {"lr": 0.001, "use_dropout": False, "use_weight_decay": False},
                0.9820157289505005,
            ),
            (
                {
                    "lr": 0.01,
                    "use_dropout": False,
                    "use_weight_decay": True,
                    "weight_decay": 0.01,
                },
                0.9627381563186646,
            ),
            (
                {
                    "lr": 0.1,
                    "use_dropout": True,
                    "dropout": 0.1,
                    "use_weight_decay": False,
                },
                0.103008933365345,
            ),
            (
                {
                    "lr": 0.01,
                    "use_dropout": True,
                    "dropout": 0.5,
                    "use_weight_decay": True,
                    "weight_decay": 0.01,
                },
                0.9476024508476257,
            ),
        ]

        benchmark = get_mnist_surrogate_benchmark(num_trials=1)

        for params, target_value in cases:
            self.assertAlmostEqual(
                benchmark.test_function.evaluate_true(params).item(),
                target_value,
            )

    def test_benchmark_creation(self) -> None:
        benchmark = get_mnist_surrogate_benchmark(num_trials=1)

        objective = benchmark.optimization_config.objective
        metric = objective.metric

        self.assertEqual(metric.name, "MNIST Test Accuracy")
        self.assertFalse(objective.minimize)
        self.assertFalse(metric.lower_is_better)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

# flake8: noqa F401
from ax.adapter import transforms
from ax.adapter.base import Adapter
from ax.adapter.factory import get_factorial, get_sobol, get_thompson
from ax.adapter.registry import Generators
from ax.adapter.torch import TorchAdapter

__all__ = [
    "Adapter",
    "Generators",
    "TorchAdapter",
    "get_factorial",
    "get_sobol",
    "get_thompson",
    "transforms",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.adapter.base import Adapter

from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCard
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import override


class MetricSummary(Analysis):
    """
    Creates a dataframe with information about each metric in the
    experiment. The resulting dataframe has one row per metric, and the
    following columns:
        - Name: the name of the metric.
        - Type: the metric subclass (e.g., Metric, BraninMetric).
        - Goal: the goal for this for this metric, based on the optimization
            config (minimize, maximize, constraint or track).
        - Bound: the bound of this metric (e.g., "<=10.0") if it is being used
            as part of an ObjectiveThreshold or OutcomeConstraint.
        - Lower is Better: whether the user prefers this metric to be lower,
            if provided.
    """

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCard:
        if experiment is None:
            raise UserInputError(
                "`MetricSummary` analysis requires an `Experiment` input"
            )
        return self._create_analysis_card(
            title=f"MetricSummary for `{experiment.name}`",
            subtitle="High-level summary of the `Metric`-s in this `Experiment`",
            df=experiment.metric_config_summary_df,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.benchmark.problems.surrogate.hss.cifar10_surrogate import (
    get_cifar10_surrogate_benchmark,
)
from ax.utils.common.testutils import TestCase


class TestCIFAR10Surrogate(TestCase):
    def test_benchmark_creation(self) -> None:
        benchmark = get_cifar10_surrogate_benchmark(num_trials=1)
        objective = benchmark.optimization_config.objective

        self.assertFalse(objective.minimize)
        self.assertFalse(objective.metric.lower_is_better)
        self.assertEqual(objective.metric.name, "CIFAR10 Test Accuracy")

    def test_cifar10_surrogate(self) -> None:
        cases = [
            (
                {
                    "lr": 0.001,
                    "two-more-blocks": False,
                    "use_softplus_activation": False,
                    "use_weight_decay": False,
                },
                0.7585159540176392,
            ),
            (
                {
                    "lr": 0.01,
                    "two-more-blocks": False,
                    "use_softplus_activation": False,
                    "use_weight_decay": True,
                    "weight_decay": 0.0001,
                },
                0.7298734784126282,
            ),
            (
                {
                    "lr": 1e-05,
                    "two-more-blocks": True,
                    "use_softplus_activation": True,
                    "softplus_beta": 2.0,
                    "use_weight_decay": False,
                },
                0.713783860206604,
            ),
            (
                {
                    "lr": 0.0001,
                    "two-more-blocks": True,
                    "use_softplus_activation": True,
                    "softplus_beta": 2.0,
                    "use_weight_decay": True,
                    "weight_decay": 1e-08,
                },
                0.7651293277740479,
            ),
        ]

        benchmark = get_cifar10_surrogate_benchmark(num_trials=1)

        for params, target_value in cases:
            self.assertAlmostEqual(
                benchmark.test_function.evaluate_true(params).item(),
                target_value,
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass

from ax.core.types import TParamValue
from torch import Tensor


@dataclass(kw_only=True)
class BenchmarkTestFunction(ABC):
    """
    The basic Ax class for generating deterministic data to benchmark against.

    (Noise - if desired - is added by the runner.)

    Args:
        outcome_names: Names of the outcomes.
        n_steps: Number of data points produced per metric and per evaluation. 1
            if data is not time-series. If data is time-series, this will
            eventually become the number of values on a `MapMetric` for
            evaluations that run to completion.
    """

    outcome_names: Sequence[str]
    n_steps: int = 1

    @abstractmethod
    def evaluate_true(self, params: Mapping[str, TParamValue]) -> Tensor:
        """
        Evaluate noiselessly.

        Returns:
            A 2d tensor of shape (len(self.outcome_names), self.n_steps).
        """
        ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field

import torch
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from botorch.test_functions.synthetic import BaseTestProblem, ConstrainedBaseTestProblem
from botorch.utils.transforms import normalize, unnormalize
from pyre_extensions import assert_is_instance
from torch import Tensor


@dataclass(kw_only=True)
class BoTorchTestFunction(BenchmarkTestFunction):
    """
    Class for generating data from a BoTorch ``BaseTestProblem``.

    Args:
        outcome_names: Names of outcomes. Should have the same length as the
            dimension of the test function, including constraints.
        botorch_problem: The BoTorch ``BaseTestProblem``.
        use_shifted_function: Whether to use the shifted version of the test function.
            If True, an offset tensor is randomly drawn from the test problem bounds,
            and the we evaluate `f(X-offset)` rather than `f(X)`. This is useful for
            changing the location of the optima for test functions that favor the
            center of the search space.
        modified_bounds: The bounds that are used by the Ax search space
            while optimizing the problem. If different from the bounds of the
            test problem, we project the parameters into the test problem
            bounds before evaluating the test problem.
            For example, if the test problem is defined on [0, 1] but the Ax
            search space is integers in [0, 10], an Ax parameter value of
            5 will correspond to 0.5 while evaluating the test problem.
            If modified bounds are not provided, the test problem will be
            evaluated using the raw parameter values.
        dummy_param_names: Names of parameters that do not affect the value of
            the test function because they are not passed to
            ``self.botorch_problem``.
            self.botorch_problem.dim + len(dummy_param_names) should equal the
            number of parameters in the ``params`` passed to ``evaluate_true``.
        n_steps: Number of data points produced per metric and per evaluation. 1
            if data is not time-series. If data is time-series, this will
            eventually become the number of values on a `MapMetric` for
            evaluations that run to completion.
    """

    outcome_names: Sequence[str]
    botorch_problem: BaseTestProblem
    use_shifted_function: bool = False
    modified_bounds: Sequence[tuple[float, float]] | None = None
    _offset: torch.Tensor | None = None
    _original_bounds: torch.Tensor | None = None
    dummy_param_names: set[str] = field(default_factory=set)

    def __post_init__(self) -> None:
        if (
            self.botorch_problem.noise_std is not None
            or getattr(self.botorch_problem, "constraint_noise_std", None) is not None
        ):
            raise ValueError(
                "noise should be set on the `BenchmarkRunner`, not the test function."
            )
        self.botorch_problem = self.botorch_problem.to(dtype=torch.double)
        self._original_bounds = self.botorch_problem.bounds.clone()
        if self.use_shifted_function:
            lo, hi = self.botorch_problem.bounds.unbind(dim=0)
            self._offset = lo + (hi - lo) * torch.rand(
                self.botorch_problem.dim, dtype=torch.double
            )
            # Modify the bounds to match the shifted problem.
            # The Ax-specified bounds are [2 * lo, 2 * hi] which implies that
            # X - offset is in [2 * lo - offset, 2 * hi - offset].
            self.botorch_problem._bounds = [
                (2 * bound[0] - offset, 2 * bound[1] - offset)
                for bound, offset in zip(self.botorch_problem._bounds, self._offset)
            ]
            self.botorch_problem.bounds = (
                2.0 * self.botorch_problem.bounds
                - assert_is_instance(self._offset, Tensor)
            )

    def tensorize_params(self, params: Mapping[str, int | float]) -> torch.Tensor:
        """Converts parameters to a 1d tensor.

        If modified bounds are provided, we normalize the parameters from the modified
        bounds to the unit cube, and then unnormalize to the original problem bounds.

        If `use_shifted_function=True`, we subtract the offset from the resulting tensor
        before returning it.
        """
        X = torch.tensor(
            [v for k, v in params.items() if k not in self.dummy_param_names],
            dtype=torch.double,
        )

        if self.modified_bounds is not None:
            # Normalize from modified bounds to unit cube.
            unit_X = normalize(
                X, torch.tensor(self.modified_bounds, dtype=torch.double).T
            )
            # Unnormalize from unit cube to original problem bounds.
            X = unnormalize(unit_X, assert_is_instance(self._original_bounds, Tensor))
        if self.use_shifted_function:
            X = X - assert_is_instance(self._offset, torch.Tensor)
        return X

    # pyre-fixme [14]: inconsistent override
    def evaluate_true(self, params: Mapping[str, float | int]) -> torch.Tensor:
        expected_n_dims = self.botorch_problem.dim + len(self.dummy_param_names)
        if len(params) != expected_n_dims:
            raise ValueError(
                f"Expected {expected_n_dims} parameters, got {len(params)}."
            )
        x = self.tensorize_params(params=params)
        # self.botorch_problem(x) has shape [n_metrics] if n_metrics > 1,
        # otherwise []. So `objectives` has shape [n_metrics]
        objectives = torch.atleast_1d(self.botorch_problem(x))
        if isinstance(self.botorch_problem, ConstrainedBaseTestProblem):
            constraints = torch.atleast_1d(self.botorch_problem.evaluate_slack_true(x))
            metrics = torch.cat([objectives, constraints], dim=-1)
        else:
            metrics = objectives
        # shape (n_metrics, 1)
        metrics = metrics.unsqueeze(1)
        if self.n_steps == 1:
            return metrics
        # shape (n_metrics, n_steps)
        return metrics.repeat(1, self.n_steps)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import json
import os
import time
from pathlib import Path

import papermill
from nbclient.exceptions import CellTimeoutError

TUTORIALS_TO_SKIP = [
    "raytune_pytorch_cnn",  # TODO: Times out CI but passes locally. Investigate.
]


def run_script(
    tutorial: Path, timeout_minutes: int, env: dict[str, str] | None = None
) -> None:
    if env is not None:
        os.environ.update(env)
    papermill.execute_notebook(
        tutorial,
        tutorial,
        # This timeout is on cell-execution time, not on total runtime.
        execution_timeout=timeout_minutes * 60,
    )


def run_tutorials(
    repo_dir: str,
    name: str | None = None,
    smoke_test: bool = False,
) -> None:
    """Run Jupyter notebooks.

    We check in the tutorial notebook un-run, and run them in CI as integration tests.
    """
    has_errors = False

    with open(os.path.join(repo_dir, "website", "tutorials.json")) as infile:
        tutorial_config = json.loads(infile.read())
    # flatten config dict
    tutorial_configs = [
        config for category in tutorial_config.values() for config in category
    ]
    # Running only the tutorial described by "name"
    if name is not None:
        tutorial_configs = [d for d in tutorial_configs if d["id"] == name]
        if len(tutorial_configs) == 0:
            raise RuntimeError(f"No tutorial found with name {name}.")
    # prepare paths for converted tutorials & files
    env = {"RUNNING_IN_PAPERMILL": "True"}
    if smoke_test:
        env["SMOKE_TEST"] = "True"

    for config in tutorial_configs:
        tid = config["id"]
        tutorial_path = os.path.join(repo_dir, "tutorials", tid, f"{tid}.ipynb")

        total_time = None

        if tid in TUTORIALS_TO_SKIP:
            print(f"Skipping execution of {tid}")
            continue
        else:
            print(f"Executing tutorial {tid}")
            start_time = time.monotonic()

            # Try / catch failures for now. We will re-raise at the end.
            timeout_minutes = 15 if smoke_test else 150
            try:
                # Execute notebook.
                run_script(
                    tutorial=tutorial_path,
                    timeout_minutes=timeout_minutes,
                    env=env,
                )
                total_time = time.monotonic() - start_time
                print(
                    f"Finished executing tutorial {tid} in {total_time:.2f} seconds. "
                )
            except CellTimeoutError:
                has_errors = True
                print(
                    f"Tutorial {tid} exceeded the maximum runtime of "
                    f"{timeout_minutes} minutes."
                )
            except Exception as e:
                has_errors = True
                print(f"Encountered error running tutorial {tid}: \n {e}")

    if has_errors:
        raise Exception("There are errors in tutorials, will not continue to publish")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate JS, HTML, ipynb, and py files for tutorials."
    )
    parser.add_argument(
        "-w", "--repo_dir", metavar="path", required=True, help="Ax repo directory."
    )
    parser.add_argument(
        "-s", "--smoke", action="store_true", help="Run in smoke test mode."
    )
    parser.add_argument(
        "-n",
        "--name",
        help="Run a specific tutorial by name. The name should not include the "
        ".ipynb extension.",
    )
    args = parser.parse_args()
    run_tutorials(
        args.repo_dir,
        smoke_test=args.smoke,
        name=args.name,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any

import torch


def get_torch_test_data(
    dtype: torch.dtype = torch.float,
    cuda: bool = False,
    constant_noise: bool = True,
    task_features: list[int] | None = None,
    offset: float = 0.0,
) -> tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    list[tuple[float, float]],
    list[int],
    list[str],
    list[str],
]:
    tkwargs: dict[str, Any] = {
        "device": torch.device("cuda" if cuda else "cpu"),
        "dtype": dtype,
    }
    X = torch.tensor(
        [
            [1.0 + offset, 2.0 + offset, 3.0 + offset],
            [2.0 + offset, 3.0 + offset, 4.0 + offset],
        ],
        **tkwargs,
    )
    Y = torch.tensor([[3.0 + offset], [4.0 + offset]], **tkwargs)
    if constant_noise:
        Yvar = torch.ones(2, 1, **tkwargs)
    else:
        Yvar = torch.tensor([[0.0 + offset], [2.0 + offset]], **tkwargs)

    bounds = [
        (0.0 + offset, 1.0 + offset),
        (1.0 + offset, 4.0 + offset),
        (2.0 + offset, 5.0 + offset),
    ]
    feature_names = ["x1", "x2", "x3"]
    task_features = [] if task_features is None else task_features
    metric_names = ["y"]
    return X, Y, Yvar, bounds, task_features, feature_names, metric_names



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from copy import deepcopy

from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.metadata_to_task import MetadataToTask
from ax.core.observation import observations_from_data
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_experiment_with_observations
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance


class MetadataToTaskTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x", parameter_type=ParameterType.FLOAT, lower=1, upper=20
                )
            ]
        )
        self.experiment = get_experiment_with_observations(
            observations=[[0.0], [1.0]],
            search_space=self.search_space,
            parameterizations=[{"x": 0.0}, {"x": 1.0}],
            candidate_metadata=[
                {Keys.TASK_FEATURE_NAME.value: 0},
                {Keys.TASK_FEATURE_NAME.value: 1},
            ],
        )
        self.observations = observations_from_data(
            experiment=self.experiment, data=self.experiment.lookup_data()
        )
        self.experiment_data = extract_experiment_data(
            experiment=self.experiment, data_loader_config=DataLoaderConfig()
        )
        self.transform_config: TConfig = {"task_values": [0, 1]}
        self.t = MetadataToTask(
            observations=self.observations, config=self.transform_config
        )

    def test_Init(self) -> None:
        self.assertEqual(len(self.t._parameter_list), 1)
        p = self.t._parameter_list[0]
        # check that the parameter options are specified in a sensible manner
        # by default if the user does not specify them explicitly
        self.assertEqual(p.name, Keys.TASK_FEATURE_NAME.value)
        self.assertEqual(p.parameter_type, ParameterType.INT)
        self.assertIsInstance(p, ChoiceParameter)
        self.assertEqual(p.values, [0, 1])
        self.assertFalse(p.is_fidelity)
        self.assertTrue(p.is_task)
        self.assertEqual(p.target_value, 0)

    def test_TransformSearchSpace(self) -> None:
        ss2 = self.search_space.clone()
        ss2 = self.t.transform_search_space(ss2)

        self.assertSetEqual(
            set(ss2.parameters.keys()),
            {"x", Keys.TASK_FEATURE_NAME.value},
        )

        p = assert_is_instance(
            ss2.parameters[Keys.TASK_FEATURE_NAME.value], ChoiceParameter
        )
        self.assertEqual(p.name, Keys.TASK_FEATURE_NAME.value)
        self.assertEqual(p.parameter_type, ParameterType.INT)
        self.assertEqual(p.values, [0, 1])
        self.assertFalse(p.is_fidelity)
        self.assertTrue(p.is_task)
        self.assertEqual(p.target_value, 0)

    def test_TransformObservationFeatures(self) -> None:
        observation_features = [obs.features for obs in self.observations]
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        expected_obs_ft2 = []
        for i, obs_ft in enumerate(observation_features):
            new_obs_ft = obs_ft.clone(
                replace_parameters={
                    **obs_ft.parameters,
                    Keys.TASK_FEATURE_NAME.value: i,
                }
            )
            del new_obs_ft.metadata[Keys.TASK_FEATURE_NAME]
            expected_obs_ft2.append(new_obs_ft)
        self.assertEqual(obs_ft2, expected_obs_ft2)
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)

    def test_transform_experiment_data(self) -> None:
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        # Check that arm data now has a new column for the transformed parameter.
        expected_task_values = [0, 1]
        self.assertEqual(
            transformed_data.arm_data[Keys.TASK_FEATURE_NAME.value].tolist(),
            expected_task_values,
        )
        # Remaining columns are unchanged.
        assert_frame_equal(
            transformed_data.arm_data.drop(columns=[Keys.TASK_FEATURE_NAME.value]),
            self.experiment_data.arm_data,
        )
        # Observation data is not changed.
        assert_frame_equal(
            transformed_data.observation_data,
            self.experiment_data.observation_data,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from unittest.mock import MagicMock

from ax.exceptions.core import UnsupportedError
from ax.generators.torch.botorch_modular.optimizer_argparse import (
    BATCH_LIMIT,
    INIT_BATCH_LIMIT,
    MAX_OPT_AGG_SIZE,
    NUM_RESTARTS,
    optimizer_argparse,
    RAW_SAMPLES,
)
from ax.utils.common.testutils import TestCase
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.knowledge_gradient import (
    qKnowledgeGradient,
    qMultiFidelityKnowledgeGradient,
)


class DummyAcquisitionFunction(AcquisitionFunction):
    def __init__(self) -> None:
        return

    # pyre-fixme[14]: Inconsistent override
    # pyre-fixme[15]: Inconsistent override
    def forward(self) -> int:
        return 0


class OptimizerArgparseTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.acqf = DummyAcquisitionFunction()
        self.default_expected_options = {
            "optimize_acqf": {
                "num_restarts": NUM_RESTARTS,
                "raw_samples": RAW_SAMPLES,
                "options": {
                    "init_batch_limit": INIT_BATCH_LIMIT,
                    "batch_limit": BATCH_LIMIT,
                    "max_optimization_problem_aggregation_size": MAX_OPT_AGG_SIZE,
                },
                "sequential": True,
            },
            "optimize_acqf_discrete_local_search": {
                "num_restarts": NUM_RESTARTS,
                "raw_samples": RAW_SAMPLES,
            },
            "optimize_acqf_discrete": {},
            "optimize_acqf_mixed": {
                "num_restarts": NUM_RESTARTS,
                "raw_samples": RAW_SAMPLES,
                "options": {
                    "init_batch_limit": INIT_BATCH_LIMIT,
                    "batch_limit": BATCH_LIMIT,
                },
            },
            "optimize_acqf_mixed_alternating": {
                "num_restarts": NUM_RESTARTS,
                "raw_samples": RAW_SAMPLES,
                "options": {
                    "init_batch_limit": INIT_BATCH_LIMIT,
                    "batch_limit": BATCH_LIMIT,
                },
            },
        }

    def test_unsupported_optimizer(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "optimizer=`wishful thinking` is not supported"
        ):
            optimizer_argparse(self.acqf, optimizer="wishful thinking")

    def test_optimizer_options(self) -> None:
        # currently there is only one function in fns_to_test
        user_options = {"foo": "bar", "num_restarts": 13}
        for optimizer in [
            "optimize_acqf",
            "optimize_acqf_discrete",
            "optimize_acqf_mixed",
            "optimize_acqf_discrete_local_search",
        ]:
            with self.subTest(optimizer=optimizer):
                parsed_options = optimizer_argparse(
                    self.acqf, optimizer_options=user_options, optimizer=optimizer
                )
                self.assertDictEqual(
                    {**self.default_expected_options[optimizer], **user_options},
                    parsed_options,
                )

        # Also test sub-options.
        inner_options = {"batch_limit": 10, "maxiter": 20}
        options = {"options": inner_options}
        for optimizer in [
            "optimize_acqf",
            "optimize_acqf_mixed",
            "optimize_acqf_mixed_alternating",
        ]:
            default = self.default_expected_options[optimizer]
            parsed_options = optimizer_argparse(
                self.acqf, optimizer_options=options, optimizer=optimizer
            )
            expected_options = {k: v for k, v in default.items() if k != "options"}
            if "options" in default:
                expected_options["options"] = {
                    **default["options"],
                    **inner_options,
                }
            else:
                expected_options["options"] = inner_options
            self.assertDictEqual(expected_options, parsed_options)

        # Error out if options is specified for an optimizer that does
        # not support the arg.
        for optimizer in [
            "optimize_acqf_discrete",
            "optimize_acqf_discrete_local_search",
        ]:
            with self.assertRaisesRegex(UnsupportedError, "`options` argument"):
                optimizer_argparse(
                    self.acqf,
                    optimizer_options={"options": {"batch_limit": 10, "maxiter": 20}},
                    optimizer=optimizer,
                )

    def test_kg(self) -> None:
        user_options = {"foo": "bar", "num_restarts": 114}
        generic_options = optimizer_argparse(
            self.acqf, optimizer_options=user_options, optimizer="optimize_acqf"
        )
        for acqf in (
            qKnowledgeGradient(model=MagicMock(), posterior_transform=MagicMock()),
            qMultiFidelityKnowledgeGradient(
                model=MagicMock(), posterior_transform=MagicMock()
            ),
        ):
            with self.subTest(acqf=acqf):
                options = optimizer_argparse(
                    acqf, optimizer_options=user_options, optimizer="optimize_acqf"
                )
                self.assertEqual(options, generic_options)

                with self.assertRaisesRegex(
                    RuntimeError,
                    "Ax is attempting to use a discrete or mixed optimizer, "
                    "`optimize_acqf_mixed`, ",
                ):
                    optimizer_argparse(acqf, optimizer="optimize_acqf_mixed")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from collections.abc import Iterator
from dataclasses import dataclass, field
from typing import Any

import numpy as np
import pandas as pd
import torch
from ax.adapter.torch import TorchAdapter
from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.benchmark_metric import (
    BenchmarkMapMetric,
    BenchmarkMapUnavailableWhileRunningMetric,
    BenchmarkMetric,
    BenchmarkTimeVaryingMetric,
)
from ax.benchmark.benchmark_problem import (
    BenchmarkProblem,
    get_moo_opt_config,
    get_soo_opt_config,
)
from ax.benchmark.benchmark_result import AggregatedBenchmarkResult, BenchmarkResult
from ax.benchmark.benchmark_step_runtime_function import TBenchmarkStepRuntimeFunction
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.benchmark.benchmark_test_functions.surrogate import SurrogateTestFunction
from ax.benchmark.benchmark_test_functions.synthetic import IdentityTestFunction
from ax.benchmark.problems.surrogate.lcbench.data import LCBenchData
from ax.benchmark.problems.surrogate.lcbench.utils import get_lcbench_parameters
from ax.benchmark.problems.synthetic.from_botorch import create_problem_from_botorch
from ax.benchmark.problems.synthetic.hss.jenatton import get_jenatton_search_space
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.parameter import ChoiceParameter, ParameterType
from ax.core.search_space import SearchSpace
from ax.core.trial import Trial
from ax.core.types import TParameterization, TParamValue
from ax.early_stopping.strategies.base import BaseEarlyStoppingStrategy
from ax.generation_strategy.external_generation_node import ExternalGenerationNode
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
)
from botorch.test_functions.multi_objective import BraninCurrin
from botorch.test_functions.synthetic import Branin


def get_single_objective_benchmark_problem(
    observe_noise_sd: bool = False,
    num_trials: int = 4,
    test_problem_kwargs: dict[str, Any] | None = None,
    report_inference_value_as_trace: bool = False,
    noise_std: float | list[float] = 0.0,
    status_quo_params: TParameterization | None = None,
) -> BenchmarkProblem:
    return create_problem_from_botorch(
        test_problem_class=Branin,
        test_problem_kwargs=test_problem_kwargs or {},
        num_trials=num_trials,
        observe_noise_sd=observe_noise_sd,
        report_inference_value_as_trace=report_inference_value_as_trace,
        noise_std=noise_std,
        status_quo_params=status_quo_params,
        baseline_value=3,
    )


def get_multi_objective_benchmark_problem(
    observe_noise_sd: bool = False,
    num_trials: int = 4,
    test_problem_class: type[BraninCurrin] = BraninCurrin,
    report_inference_value_as_trace: bool = False,
) -> BenchmarkProblem:
    return create_problem_from_botorch(
        test_problem_class=test_problem_class,
        test_problem_kwargs={},
        num_trials=num_trials,
        observe_noise_sd=observe_noise_sd,
        report_inference_value_as_trace=report_inference_value_as_trace,
        baseline_value=0.0,
    )


def get_soo_surrogate_test_function(lazy: bool = True) -> SurrogateTestFunction:
    experiment = get_branin_experiment(with_completed_trial=True)
    surrogate = TorchAdapter(
        experiment=experiment,
        search_space=experiment.search_space,
        generator=BoTorchGenerator(),
        data=experiment.lookup_data(),
        transforms=[],
    )
    if lazy:
        test_function = SurrogateTestFunction(
            outcome_names=["branin"], name="test", get_surrogate=lambda: surrogate
        )
    else:
        test_function = SurrogateTestFunction(
            outcome_names=["branin"],
            name="test",
            _surrogate=surrogate,
        )
    return test_function


def get_soo_surrogate() -> BenchmarkProblem:
    experiment = get_branin_experiment(with_completed_trial=True)
    test_function = get_soo_surrogate_test_function()

    optimization_config = get_soo_opt_config(
        outcome_names=test_function.outcome_names,
        observe_noise_sd=True,
    )

    return BenchmarkProblem(
        name="test",
        search_space=experiment.search_space,
        optimization_config=optimization_config,
        num_trials=6,
        optimal_value=0.0,
        baseline_value=3.0,
        test_function=test_function,
    )


def get_moo_surrogate() -> BenchmarkProblem:
    experiment = get_branin_experiment_with_multi_objective(with_completed_trial=True)
    surrogate = TorchAdapter(
        experiment=experiment,
        search_space=experiment.search_space,
        generator=BoTorchGenerator(),
        data=experiment.lookup_data(),
        transforms=[],
    )

    outcome_names = ["branin_a", "branin_b"]
    test_function = SurrogateTestFunction(
        name="test", outcome_names=outcome_names, get_surrogate=lambda: surrogate
    )
    optimization_config = get_moo_opt_config(
        outcome_names=outcome_names,
        ref_point=[0.0, 0.0],
        observe_noise_sd=True,
    )

    return BenchmarkProblem(
        name="test",
        search_space=experiment.search_space,
        optimization_config=optimization_config,
        num_trials=10,
        optimal_value=1.0,
        baseline_value=0.0,
        test_function=test_function,
    )


def get_benchmark_result() -> BenchmarkResult:
    problem = get_single_objective_benchmark_problem()

    return BenchmarkResult(
        name="test_benchmarking_result",
        seed=0,
        experiment=Experiment(
            name="test_benchmarking_experiment",
            search_space=problem.search_space,
            optimization_config=problem.optimization_config,
            is_test=True,
        ),
        inference_trace=np.ones(4),
        oracle_trace=np.zeros(4),
        cost_trace=np.zeros(4),
        optimization_trace=np.array([3, 2, 1, 0.1]),
        score_trace=np.array([3, 2, 1, 0.1]),
        fit_time=0.1,
        gen_time=0.2,
    )


def get_aggregated_benchmark_result() -> AggregatedBenchmarkResult:
    result = get_benchmark_result()
    return AggregatedBenchmarkResult.from_benchmark_results([result, result])


@dataclass(kw_only=True)
class DummyTestFunction(BenchmarkTestFunction):
    outcome_names: list[str] = field(default_factory=list)
    num_outcomes: int = 1
    dim: int = 6

    def __post_init__(self) -> None:
        self.outcome_names = [f"objective_{i}" for i in range(self.num_outcomes)]

    # pyre-fixme[14]: Inconsistent override, as dict[str, float] is not a
    # `TParameterization`
    def evaluate_true(self, params: dict[str, float]) -> torch.Tensor:
        value = sum(elt**2 for elt in params.values())
        return value * torch.ones(self.num_outcomes, dtype=torch.double)


def get_jenatton_arm(i: int) -> Arm:
    """
    Args:
        i Non-negative int.
    """
    jenatton_x_params = {f"x{j}": j % (i + 1) for j in range(1, 8)}
    jenatton_r_params = {"r8": 0.0, "r9": 0.0}
    return Arm(parameters={**jenatton_x_params, **jenatton_r_params}, name=f"0_{i}")


def get_jenatton_experiment() -> Experiment:
    experiment = Experiment(
        search_space=get_jenatton_search_space(),
        name="test_jenatton",
        is_test=True,
    )
    return experiment


def get_jenatton_trials(n_trials: int) -> dict[int, Trial]:
    experiment = get_jenatton_experiment()
    for i in range(n_trials):
        trial = experiment.new_trial()
        trial.add_arm(get_jenatton_arm(i=i))
    # pyre-fixme: Incompatible return type [7]: Expected `Dict[int, Trial]` but
    # got `Dict[int, BaseTrial]`.
    return experiment.trials


def get_jenatton_batch_trial() -> BatchTrial:
    experiment = get_jenatton_experiment()
    trial = experiment.new_batch_trial()
    trial.add_arm(get_jenatton_arm(0))
    trial.add_arm(get_jenatton_arm(1))
    return trial


class DeterministicGenerationNode(ExternalGenerationNode):
    """
    A GenerationNode that explores a discrete search space with one parameter
    deterministically.
    """

    def __init__(
        self,
        search_space: SearchSpace,
    ) -> None:
        if len(search_space.parameters) != 1:
            raise ValueError(
                "DeterministicGenerationNode only supports search spaces with one "
                "parameter."
            )
        param = list(search_space.parameters.values())[0]
        if not isinstance(param, ChoiceParameter):
            raise ValueError(
                "DeterministicGenerationNode only supports ChoiceParameters."
            )
        super().__init__(node_name="Deterministic")

        self.param_name: str = param.name
        self.iterator: Iterator[TParamValue] = iter(param.values)

    def update_generator_state(self, experiment: Experiment, data: Data) -> None:
        return

    def get_next_candidate(
        self, pending_parameters: list[TParameterization]
    ) -> TParameterization:
        return {self.param_name: next(self.iterator)}


def get_discrete_search_space(n_values: int = 20) -> SearchSpace:
    return SearchSpace(
        parameters=[
            ChoiceParameter(
                name="x0",
                parameter_type=ParameterType.INT,
                values=list(range(n_values)),
            )
        ]
    )


def get_async_benchmark_method(
    early_stopping_strategy: BaseEarlyStoppingStrategy | None = None,
    max_pending_trials: int = 2,
) -> BenchmarkMethod:
    gs = GenerationStrategy(
        nodes=[DeterministicGenerationNode(search_space=get_discrete_search_space())]
    )
    return BenchmarkMethod(
        generation_strategy=gs,
        distribute_replications=False,
        max_pending_trials=max_pending_trials,
        batch_size=1,
        early_stopping_strategy=early_stopping_strategy,
    )


def get_async_benchmark_problem(
    map_data: bool,
    step_runtime_fn: TBenchmarkStepRuntimeFunction | None = None,
    n_steps: int = 1,
    lower_is_better: bool = False,
) -> BenchmarkProblem:
    search_space = get_discrete_search_space()
    test_function = IdentityTestFunction(n_steps=n_steps)
    optimization_config = get_soo_opt_config(
        outcome_names=["objective"],
        use_map_metric=map_data,
        observe_noise_sd=True,
        lower_is_better=lower_is_better,
    )

    return BenchmarkProblem(
        name="test",
        search_space=search_space,
        optimization_config=optimization_config,
        test_function=test_function,
        num_trials=4,
        baseline_value=19 if lower_is_better else 0,
        optimal_value=0 if lower_is_better else 19,
        step_runtime_function=step_runtime_fn,
    )


def get_benchmark_metric() -> BenchmarkMetric:
    return BenchmarkMetric(name="test", lower_is_better=True)


def get_benchmark_map_metric() -> BenchmarkMapMetric:
    return BenchmarkMapMetric(name="test", lower_is_better=True)


def get_benchmark_time_varying_metric() -> BenchmarkTimeVaryingMetric:
    return BenchmarkTimeVaryingMetric(name="test", lower_is_better=True)


def get_benchmark_map_unavailable_while_running_metric() -> (
    BenchmarkMapUnavailableWhileRunningMetric
):
    return BenchmarkMapUnavailableWhileRunningMetric(name="test", lower_is_better=True)


def get_mock_lcbench_data() -> LCBenchData:
    """
    Used for mocking out `load_lcbench_data` to avoid downloading data from the
    internet.
    """
    timestamp_series = pd.Series([0], index=pd.Index([0], name="trial"))
    parameters = get_lcbench_parameters()

    parameter_df = pd.DataFrame(
        {name: [param.lower, param.upper] for name, param in parameters.items()}
    )
    metric_series = pd.Series(
        [0],
        index=pd.MultiIndex.from_tuples([(0, 0)], names=["trial", "epoch"]),
    )

    return LCBenchData(
        parameter_df=parameter_df,
        metric_series=metric_series,
        timestamp_series=timestamp_series,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from ax.utils.common.string_utils import sanitize_name
from ax.utils.common.testutils import TestCase


class StringUtilsTest(TestCase):
    def test_sanitize_name(self) -> None:
        self.assertEqual(sanitize_name("foo.bar.baz"), "foo__dot__bar__dot__baz")
        self.assertEqual(
            sanitize_name("foo.bar/11:Baz"), "foo__dot__bar__slash__11__colon__Baz"
        )
        self.assertEqual(
            sanitize_name("foo.bar + 0.1 * baz"), "foo__dot__bar + 0.1 * baz"
        )
        for s in ("foo;", "foo\\", "'foo"):
            with self.assertRaisesRegex(
                ValueError, "has forbidden control characters."
            ):
                sanitize_name(s=s)
        self.assertEqual(sanitize_name(s="foo"), "foo")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections.abc import Sequence
from logging import Logger
from typing import Any

import torch
from ax.exceptions.core import UserInputError
from ax.generators.torch.botorch_modular.kernels import (
    DefaultMaternKernel,
    DefaultRBFKernel,
    ScaleMaternKernel,
)
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import _argparse_type_encoder
from botorch.models import MultiTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.model import Model
from botorch.utils.datasets import MultiTaskDataset, SupervisedDataset
from botorch.utils.dispatcher import Dispatcher
from botorch.utils.transforms import normalize_indices
from botorch.utils.types import _DefaultType, DEFAULT
from gpytorch.kernels.kernel import Kernel
from gpytorch.kernels.linear_kernel import LinearKernel
from gpytorch.priors.torch_priors import Prior
from pyre_extensions import assert_is_instance, none_throws

logger: Logger = get_logger(__name__)

covar_module_argparse = Dispatcher(
    name="covar_module_argparse", encoder=_argparse_type_encoder
)


@covar_module_argparse.register(Kernel)
def _covar_module_argparse_base(
    covar_module_class: type[Kernel],
    botorch_model_class: type[Model] | None = None,
    dataset: SupervisedDataset | None = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """
    Extract the covar module kwargs form the given arguments.

    Args:
        covar_module_class: Covariance module class.
        botorch_model_class: ``Model`` class to be used as the underlying
            BoTorch model.
        dataset: Dataset containing feature matrix and the response.

    Returns:
        A dictionary with covar module kwargs.
    """
    kwargs.pop("remove_task_features", None)
    return {**kwargs}


@covar_module_argparse.register(ScaleMaternKernel)
def _covar_module_argparse_scale_matern(
    covar_module_class: type[ScaleMaternKernel],
    botorch_model_class: type[Model],
    dataset: SupervisedDataset,
    ard_num_dims: int | _DefaultType = DEFAULT,
    batch_shape: torch.Size | _DefaultType = DEFAULT,
    lengthscale_prior: Prior | None = None,
    outputscale_prior: Prior | None = None,
    remove_task_features: bool = False,
    **kwargs: Any,
) -> dict[str, Any]:
    """Extract the base covar module kwargs form the given arguments.

    NOTE: This setup does not allow for setting multi-dimensional priors,
        with different priors over lengthscales.

    Args:
        covar_module_class: Covariance module class.
        botorch_model_class: Model class to be used as the underlying
            BoTorch model.
        dataset: Dataset containing feature matrix and the response.
        ard_num_dims: Number of lengthscales per feature.
        batch_shape: The number of lengthscales per batch.
        lengthscale_prior: Lenthscale prior.
        outputscale_prior: Outputscale prior.
        remove_task_features: A boolean indicating whether to remove the task
            features (e.g. when using a SingleTask model on a MultiTaskDataset).

    Returns:
        A dictionary with covar module kwargs.
    """
    ard_num_dims, batch_shape, active_dims = _get_default_ard_num_dims_and_batch_shape(
        ard_num_dims=ard_num_dims,
        batch_shape=batch_shape,
        botorch_model_class=botorch_model_class,
        dataset=dataset,
        remove_task_features=remove_task_features,
    )
    return _covar_module_argparse_base(
        covar_module_class=covar_module_class,
        dataset=dataset,
        botorch_model_class=botorch_model_class,
        ard_num_dims=ard_num_dims,
        lengthscale_prior=lengthscale_prior,
        outputscale_prior=outputscale_prior,
        batch_shape=batch_shape,
        active_dims=active_dims,
        **kwargs,
    )


@covar_module_argparse.register(LinearKernel)
def _covar_module_argparse_linear(
    covar_module_class: type[LinearKernel],
    botorch_model_class: type[Model],
    dataset: SupervisedDataset,
    ard_num_dims: int | _DefaultType = DEFAULT,
    active_dims: Sequence[int] | None = None,
    batch_shape: torch.Size | _DefaultType = DEFAULT,
    variance_prior: Prior | None = None,
    remove_task_features: bool = False,
    **kwargs: Any,
) -> dict[str, Any]:
    """Extract the covar module kwargs for a LinearKernle from the given arguments.

    Args:
        covar_module_class: Covariance module class.
        botorch_model_class: Model class to be used as the underlying
            BoTorch model.
        dataset: Dataset containing feature matrix and the response.
        ard_num_dims: Number of lengthscales per feature.
        active_dims: The active dimensions of the kernel.
        batch_shape: The number of lengthscales per batch.
        variance_prior: Variance prior.
        remove_task_features: A boolean indicating whether to exclude the task
            feature in the kernel.

    Returns:
        A dictionary with covar module kwargs.
    """
    ard_num_dims, batch_shape, active_dims = _get_default_ard_num_dims_and_batch_shape(
        ard_num_dims=ard_num_dims,
        batch_shape=batch_shape,
        botorch_model_class=botorch_model_class,
        dataset=dataset,
        remove_task_features=remove_task_features,
        active_dims=active_dims,
    )
    return _covar_module_argparse_base(
        covar_module_class=covar_module_class,
        dataset=dataset,
        botorch_model_class=botorch_model_class,
        ard_num_dims=ard_num_dims,
        variance_prior=variance_prior,
        batch_shape=batch_shape,
        active_dims=active_dims,
        **kwargs,
    )


@covar_module_argparse.register((DefaultRBFKernel, DefaultMaternKernel))
def _covar_module_argparse_default_rbf(
    covar_module_class: type[DefaultRBFKernel] | type[DefaultMaternKernel],
    botorch_model_class: type[Model],
    dataset: SupervisedDataset,
    ard_num_dims: int | _DefaultType = DEFAULT,
    inactive_features: Sequence[str] | None = None,
    active_dims: Sequence[int] | None = None,
    batch_shape: torch.Size | _DefaultType = DEFAULT,
    remove_task_features: bool = False,
    **kwargs: Any,
) -> dict[str, Any]:
    """Constructs inputs for ``DefaultRBFKernel``.

    The key feature of this helper is the support for ``inactive_features`` input,
    which is used to determine the ``active_dims`` input for the kernel by removing
    the dimensions that correspond to the given feature names in the dataset.
    """
    if inactive_features is not None and active_dims is not None:
        raise UserInputError(
            "Only one of `active_dims` or `inactive_features` inputs can be provided. "
            f"Got {active_dims=}, {inactive_features=}."
        )
    d = len(dataset.feature_names)
    if inactive_features is not None:
        all_dims = set(range(d))
        inactive_dims = [dataset.feature_names.index(ft) for ft in inactive_features]
        active_dims = list(all_dims.difference(inactive_dims))
    ard_num_dims, batch_shape, active_dims = _get_default_ard_num_dims_and_batch_shape(
        ard_num_dims=ard_num_dims,
        batch_shape=batch_shape,
        botorch_model_class=botorch_model_class,
        dataset=dataset,
        remove_task_features=remove_task_features,
        active_dims=active_dims,
    )
    if ard_num_dims is DEFAULT:
        ard_num_dims = len(active_dims) if active_dims is not None else d
    return _covar_module_argparse_base(
        covar_module_class=covar_module_class,
        dataset=dataset,
        botorch_model_class=botorch_model_class,
        ard_num_dims=ard_num_dims,
        batch_shape=batch_shape,
        active_dims=active_dims,
        **kwargs,
    )


def _get_default_batch_shape(
    dataset: SupervisedDataset,
    batch_shape: torch.Size | _DefaultType = DEFAULT,
) -> torch.Size:
    if (batch_shape is DEFAULT) and (dataset.Y.shape[-1:] == torch.Size([1])):
        return torch.Size([])
    if batch_shape is DEFAULT:
        return dataset.Y.shape[-1:]
    return assert_is_instance(batch_shape, torch.Size)


def _get_default_ard_num_dims_and_batch_shape(
    ard_num_dims: int | _DefaultType,
    batch_shape: torch.Size | _DefaultType,
    botorch_model_class: type[Model],
    dataset: SupervisedDataset,
    remove_task_features: bool,
    active_dims: Sequence[int] | None = None,
) -> tuple[int | _DefaultType, torch.Size | _DefaultType, list[int] | None]:
    """Helper method for constructing shared inputs across kernels.

    Args:
        ard_num_dims: The number of lengthscales.
        batch_shape: The batch shape of the kernel.
        botorch_model_class: The BoTorch model class.
        dataset: The dataset.
        remove_task_features: A boolean indicating whether to exclude the task
            feature in the kernel.
        active_dims: The active dimensions of the kernel.

    Returns:
        A tuple of (ard_num_dims, batch_shape, active_dims).
    """
    if active_dims is not None:
        active_dims = none_throws(
            normalize_indices(indices=list(active_dims), d=len(dataset.feature_names))
        )
        num_active_dims = len(active_dims)
    else:
        num_active_dims = dataset.X.shape[-1]
    if issubclass(botorch_model_class, MultiTaskGP):
        if ard_num_dims is DEFAULT:
            ard_num_dims = num_active_dims - 1
        if batch_shape is DEFAULT:
            batch_shape = torch.Size([])
    if issubclass(botorch_model_class, SingleTaskGP):
        if ard_num_dims is DEFAULT:
            ard_num_dims = num_active_dims
            if remove_task_features:
                if isinstance(dataset, MultiTaskDataset):
                    if active_dims is None:
                        active_dims = list(range(dataset.X.shape[-1]))
                    logger.debug(
                        "Excluding task feature from covar_module.", stacklevel=6
                    )
                    task_feature_index = (
                        -1
                        if dataset.task_feature_index is None
                        else dataset.task_feature_index
                    )
                    normalized_task_idx = none_throws(
                        normalize_indices(
                            indices=[task_feature_index],
                            d=dataset.X.shape[-1],
                        )
                    )[0]
                    ard_num_dims -= 1
                    active_dims = [i for i in active_dims if i != normalized_task_idx]
        batch_shape = _get_default_batch_shape(dataset=dataset, batch_shape=batch_shape)

    return ard_num_dims, batch_shape, active_dims



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import pandas as pd
from ax.adapter.factory import get_sobol
from ax.adapter.registry import Generators

from ax.analysis.healthcheck.constraints_feasibility import (
    ConstraintsFeasibilityAnalysis,
)
from ax.analysis.healthcheck.healthcheck_analysis import HealthcheckStatus
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.optimization_config import OptimizationConfig
from ax.generation_strategy.generation_node import GenerationNode
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
)
from ax.utils.testing.mock import mock_botorch_optimize
from pyre_extensions import assert_is_instance


class TestConstraintsFeasibilityAnalysis(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        experiment = get_branin_experiment_with_multi_objective(
            with_batch=False,
            with_status_quo=True,
            with_relative_constraint=True,
        )
        self.df_metric_a = pd.DataFrame(
            {
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"],
                "metric_name": ["branin_a"] * 6,
                "mean": list(np.random.normal(0, 1, 6)),
                "sem": [0.1] * 6,
                "trial_index": [0] * 6,
            }
        )
        self.df_metric_b = pd.DataFrame(
            {
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"],
                "metric_name": ["branin_b"] * 6,
                "mean": list(np.random.normal(0, 1, 6)),
                "sem": [0.1] * 6,
                "trial_index": [0] * 6,
            }
        )
        self.df_metric_d = pd.DataFrame(
            {
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"],
                "metric_name": ["branin_d"] * 6,
                "mean": [1.0, 1.0, 2.0, 3.0, 4.0, 5.0],
                "sem": [0.1] * 6,
                "trial_index": [0] * 6,
            }
        )
        df = pd.concat(
            [self.df_metric_a, self.df_metric_b, self.df_metric_d], ignore_index=True
        )

        sobol = get_sobol(search_space=experiment.search_space)
        experiment.new_batch_trial(generator_run=sobol.gen(5))

        batch_trial = assert_is_instance(experiment.trials[0], BatchTrial)

        batch_trial.add_arm(experiment.status_quo)
        batch_trial.add_status_quo_arm(weight=1.0)
        experiment.trials[0].mark_running(no_runner_required=True)
        experiment.trials[0].mark_completed()

        experiment.attach_data(data=Data(df=df))

        generation_strategy = GenerationStrategy(
            name="gs",
            nodes=[
                GenerationNode(
                    node_name="gn",
                    generator_specs=[
                        GeneratorSpec(
                            generator_enum=Generators.BOTORCH_MODULAR,
                        )
                    ],
                )
            ],
        )
        generation_strategy.experiment = experiment
        generation_strategy._curr._fit(experiment=experiment)
        self.experiment: Experiment = experiment
        self.generation_strategy: GenerationStrategy = generation_strategy

    @mock_botorch_optimize
    def test_compute(self) -> None:
        self.setUp()
        cfa = ConstraintsFeasibilityAnalysis()
        card = cfa.compute(
            experiment=self.experiment, generation_strategy=self.generation_strategy
        )
        self.assertEqual(card.name, "ConstraintsFeasibilityAnalysis")
        self.assertEqual(card.title, "Ax Constraints Feasibility Success")
        self.assertEqual(card.subtitle, "All constraints are feasible.")

        df_metric_d = pd.DataFrame(
            {
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"],
                "metric_name": ["branin_d"] * 6,
                "mean": [1.0, -1.0, -2.0, -3.0, -4.0, -5.0],
                "sem": [0.1] * 6,
                "trial_index": [0] * 6,
            }
        )
        df = pd.concat(
            [self.df_metric_a, self.df_metric_b, df_metric_d],
            ignore_index=True,
        )
        experiment = self.experiment
        generation_strategy = self.generation_strategy
        experiment.attach_data(data=Data(df=df))
        generation_strategy.experiment = experiment
        generation_strategy._curr._fit(experiment=experiment)
        cfa = ConstraintsFeasibilityAnalysis()
        card = cfa.compute(
            experiment=experiment, generation_strategy=generation_strategy
        )
        self.assertEqual(card.name, "ConstraintsFeasibilityAnalysis")
        self.assertEqual(card.title, "Ax Constraints Feasibility Warning")
        subtitle = (
            "The constraints feasibility health check utilizes "
            "samples drawn during the optimization process to assess the "
            "feasibility of constraints set on the experiment. Given these "
            "samples, the model believes there is at least a "
            "0.95 probability that the constraints will be "
            "violated. We suggest relaxing the bounds for the constraints "
            "on this Experiment."
        )
        self.assertEqual(card.subtitle, subtitle)
        self.assertEqual(card.get_status(), HealthcheckStatus.WARNING)

        # experiment with no constraints
        experiment.optimization_config = OptimizationConfig(
            objective=Objective(metric=Metric(name="branin_a"), minimize=False),
            outcome_constraints=[],
        )
        cfa = ConstraintsFeasibilityAnalysis()
        card = cfa.compute(
            experiment=experiment, generation_strategy=generation_strategy
        )
        self.assertEqual(card.name, "ConstraintsFeasibilityAnalysis")
        self.assertEqual(card.title, "Ax Constraints Feasibility Success")
        self.assertEqual(card.subtitle, "No constraints are specified.")
        self.assertEqual(card.get_status(), HealthcheckStatus.PASS)

    def test_no_optimization_config(self) -> None:
        experiment = get_branin_experiment(has_optimization_config=False)
        cfa = ConstraintsFeasibilityAnalysis()
        card = cfa.compute(experiment=experiment, generation_strategy=None)
        self.assertEqual(card.name, "ConstraintsFeasibilityAnalysis")
        self.assertEqual(card.title, "Ax Constraints Feasibility Success")
        self.assertEqual(card.subtitle, "No optimization config is specified.")
        self.assertEqual(card.get_status(), HealthcheckStatus.PASS)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from contextlib import ExitStack
from copy import deepcopy
from unittest import mock
from unittest.mock import Mock, patch

import numpy as np
from ax.adapter.base import Adapter
from ax.adapter.torch import TorchAdapter
from ax.adapter.transforms.derelativize import Derelativize, logger
from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.observation import ObservationData
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint, ScalarizedOutcomeConstraint
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import ComparisonOp
from ax.exceptions.core import DataRequiredError
from ax.generators.base import Generator
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_data,
    get_branin_experiment,
    get_experiment_with_observations,
)
from ax.utils.testing.mock import mock_botorch_optimize
from pyre_extensions import none_throws


class DerelativizeTransformTest(TestCase):
    def test_DerelativizeTransform(self) -> None:
        for negative_metrics in [False, True]:
            sq_sign = -1.0 if negative_metrics else 1.0
            observed_means = sq_sign * np.array([[1.0, 2.0], [1.0, 6.0]])
            sq_b_observed = np.mean(observed_means[1, 1])
            experiment = get_experiment_with_observations(
                observations=observed_means.tolist(),
                sems=[[1.0, 2.0], [1.0, 2.0]],
                parameterizations=[{"x": 2.0, "y": 10.0}, {"x": None, "y": None}],
            )

            predicted_means = sq_sign * np.array([3.0, 5.0])
            sq_b_predicted = predicted_means[-1]
            predict_return_value = [
                ObservationData(
                    means=predicted_means,
                    covariance=np.array([[1.0, 0.0], [0.0, 1.0]]),
                    metric_names=["m1", "m2"],
                )
            ]
            with ExitStack() as es:
                mock_predict = es.enter_context(
                    mock.patch(
                        "ax.adapter.base.Adapter._predict",
                        autospec=True,
                        return_value=predict_return_value,
                    )
                )
                self._test_DerelativizeTransform(
                    experiment=experiment,
                    mock_predict=mock_predict,
                    sq_b_observed=sq_b_observed,
                    sq_b_predicted=sq_b_predicted,
                )

    def _test_DerelativizeTransform(
        self,
        experiment: Experiment,
        mock_predict: Mock,
        sq_b_observed: float,
        sq_b_predicted: float,
    ) -> None:
        t = Derelativize(search_space=None, observations=[])

        # Adapter with in-design status quo
        search_space = SearchSpace(
            parameters=[
                RangeParameter("x", ParameterType.FLOAT, 0, 20),
                RangeParameter("y", ParameterType.FLOAT, 0, 20),
            ]
        )
        experiment.search_space = search_space
        # clone experiment since we can't overwrite sq in subsequent tests
        experiment_1 = experiment.clone_with(
            name="experiment_1",
            status_quo=Arm(parameters={"x": 2.0, "y": 10.0}, name="0_0"),
        )
        g = Adapter(experiment=experiment_1, generator=Generator())

        # Test with no relative constraints
        objective = Objective(Metric("c"), minimize=True)
        oc = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("m1"), Metric("m2")],
                    op=ComparisonOp.LEQ,
                    bound=2,
                    weights=[0.5, 0.5],
                    relative=False,
                ),
            ],
        )
        oc2 = t.transform_optimization_config(oc, g, None)
        self.assertTrue(oc == oc2)

        # Test with relative constraint, in-design status quo
        relative_bound = -10
        oc = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                OutcomeConstraint(
                    Metric("m2"), ComparisonOp.LEQ, bound=relative_bound, relative=True
                ),
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("m1"), Metric("m2")],
                    weights=[0.0, 1.0],
                    op=ComparisonOp.LEQ,
                    bound=relative_bound,
                    relative=True,
                ),
            ],
        )
        oc2 = t.transform_optimization_config(oc, g, None)
        sq_val = sq_b_predicted
        absolute_bound = (1 + np.sign(sq_val) * relative_bound / 100.0) * sq_val
        self.assertTrue(
            oc2.outcome_constraints
            == [
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                OutcomeConstraint(
                    Metric("m2"), ComparisonOp.LEQ, bound=absolute_bound, relative=False
                ),
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("m1"), Metric("m2")],
                    weights=[0.0, 1.0],
                    op=ComparisonOp.LEQ,
                    bound=absolute_bound,
                    relative=False,
                ),
            ]
        )
        obsf = mock_predict.call_args.kwargs["observation_features"][0]
        self.assertEqual(obsf.parameters, {"x": 2.0, "y": 10.0})
        self.assertEqual(mock_predict.call_count, 1)

        # The model should not be used when `use_raw_status_quo` is True
        t2 = deepcopy(t)
        t2.config["use_raw_status_quo"] = True
        t2.transform_optimization_config(deepcopy(oc), g, None)
        self.assertEqual(mock_predict.call_count, 1)

        # Test with relative constraint, out-of-design status quo
        mock_predict.side_effect = RuntimeError()
        experiment_2 = experiment.clone_with(
            name="experiment_2",
            status_quo=Arm(parameters={"x": None, "y": None}, name="1_0"),
        )
        g = Adapter(experiment=experiment_2, generator=Generator())
        oc = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                OutcomeConstraint(
                    Metric("m2"), ComparisonOp.LEQ, bound=relative_bound, relative=True
                ),
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("m1"), Metric("m2")],
                    weights=[0.0, 1.0],
                    op=ComparisonOp.LEQ,
                    bound=relative_bound,
                    relative=True,
                ),
            ],
        )
        oc2 = t.transform_optimization_config(oc, g, None)
        sq_val = sq_b_observed
        absolute_bound = (1 + np.sign(sq_val) * relative_bound / 100.0) * sq_val
        self.assertEqual(
            oc2.outcome_constraints,
            [
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                OutcomeConstraint(
                    Metric("m2"), ComparisonOp.LEQ, bound=absolute_bound, relative=False
                ),
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("m1"), Metric("m2")],
                    weights=[0.0, 1.0],
                    op=ComparisonOp.LEQ,
                    bound=absolute_bound,
                    relative=False,
                ),
            ],
        )
        self.assertEqual(mock_predict.call_count, 1)

        # Raises error if predict fails with in-design status quo
        experiment_3 = experiment.clone_with(name="experiment_3")
        experiment_3 = experiment.clone_with(
            name="experiment_3",
            status_quo=Arm(parameters={"x": 2.0, "y": 10.0}, name="0_0"),
        )
        g = Adapter(experiment=experiment_3, generator=Generator())
        oc = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                OutcomeConstraint(
                    Metric("m2"), ComparisonOp.LEQ, bound=-10, relative=True
                ),
            ],
        )
        with self.assertRaises(RuntimeError):
            t.transform_optimization_config(oc, g, None)

        # Bypasses error if use_raw_sq
        t2 = Derelativize(
            search_space=None,
            observations=[],
            config={"use_raw_status_quo": True},
        )
        t2.transform_optimization_config(deepcopy(oc), g, None)

        # But not if sq arm is not available.
        with patch(
            f"{Derelativize.__module__}.unwrap_observation_data", return_value=({}, {})
        ), self.assertRaisesRegex(
            DataRequiredError, "Status-quo metric value not yet available for metric "
        ):
            t2.transform_optimization_config(deepcopy(oc), g, None)

        # Same for scalarized constraint only.
        oc_scalarized_only = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("m1"), Metric("m2")],
                    weights=[0.0, 1.0],
                    op=ComparisonOp.LEQ,
                    bound=-10,
                    relative=True,
                ),
            ],
        )
        with patch(
            f"{Derelativize.__module__}.unwrap_observation_data", return_value=({}, {})
        ), self.assertRaisesRegex(
            DataRequiredError,
            "Status-quo metric value not yet available for metric\\(s\\) ",
        ):
            t2.transform_optimization_config(deepcopy(oc_scalarized_only), g, None)

        # Raises error with relative constraint, no status quo.
        g = Adapter(
            experiment=Experiment(search_space=search_space), generator=Generator()
        )
        with self.assertRaises(DataRequiredError):
            t.transform_optimization_config(deepcopy(oc), g, None)

        # Raises error with relative constraint, no adapter.
        with self.assertRaises(ValueError):
            t.transform_optimization_config(deepcopy(oc), None, None)

    def test_errors(self) -> None:
        t = Derelativize(
            search_space=None,
            observations=[],
        )
        oc = OptimizationConfig(
            objective=Objective(Metric("c"), minimize=False),
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("m1"), ComparisonOp.LEQ, bound=2, relative=True
                )
            ],
        )
        search_space = SearchSpace(
            parameters=[RangeParameter("x", ParameterType.FLOAT, 0, 20)]
        )
        adapter = Adapter(
            experiment=Experiment(search_space=search_space), generator=Generator()
        )
        with self.assertRaises(ValueError):
            t.transform_optimization_config(
                optimization_config=oc, adapter=None, fixed_features=None
            )
        with self.assertRaises(DataRequiredError):
            t.transform_optimization_config(
                optimization_config=oc, adapter=adapter, fixed_features=None
            )

    @mock_botorch_optimize
    def test_with_out_of_design_status_quo(self) -> None:
        # This test checks that model predictions are not used when the status quo
        # is out of design, regardless of whether it has data attached.

        # Branin experiment with out of design status quo.
        # Status quo is 0.0, so we adjust the search space to have it out of design.
        exp = get_branin_experiment(
            with_status_quo=True,
            with_completed_trial=True,
            with_relative_constraint=True,
            search_space=SearchSpace(
                parameters=[
                    RangeParameter(
                        name=name,
                        parameter_type=ParameterType.FLOAT,
                        lower=5.0,
                        upper=10.0,
                    )
                    for name in ("x1", "x2")
                ]
            ),
        )
        for attach_data_for_sq in [False, True]:
            if attach_data_for_sq:
                # Attach data for status quo, which will get filtered out.
                trial = exp.new_trial().add_arm(exp.status_quo).run().mark_completed()
                exp.attach_data(get_branin_data(trials=[trial], metrics=exp.metrics))

            for with_raw_sq in [False, True]:
                adapter = TorchAdapter(
                    experiment=exp,
                    generator=BoTorchGenerator(),
                    transforms=[Derelativize],
                    transform_configs={
                        "Derelativize": {"use_raw_status_quo": with_raw_sq}
                    },
                    expand_model_space=False,  # To keep SQ out of design.
                )
                if attach_data_for_sq:
                    # Since SQ is out of design, we shouldn't be using
                    # model predictions.
                    with mock.patch.object(adapter, "predict") as mock_predict:
                        adapter.gen(n=1)
                    # Check that it wasn't called with SQ features.
                    # It will be called after gen with generated candidate.
                    self.assertFalse(
                        any(
                            args.args[0] == [none_throws(adapter.status_quo).features]
                            for args in mock_predict.call_args_list
                        )
                    )
                else:
                    # If there is no data, SQ is not set and we get an error.
                    with self.assertRaisesRegex(
                        DataRequiredError, "was not fit with status quo"
                    ):
                        adapter.gen(n=1)

    @mock_botorch_optimize
    def test_warning_if_raw_sq_is_out_of_CI(self) -> None:
        # This test checks that a warning is raised if the raw status quo values
        # are outside of the CI for the predictions.

        exp = get_branin_experiment(
            with_status_quo=True,
            with_completed_trial=True,
            with_relative_constraint=True,
        )
        # Add data for SQ.
        trial = exp.new_trial().add_arm(exp.status_quo).run().mark_completed()
        exp.attach_data(get_branin_data(trials=[trial], metrics=exp.metrics))

        adapter = TorchAdapter(
            experiment=exp, generator=BoTorchGenerator(), transforms=[Derelativize]
        )
        # Shouldn't log in regular usage here.
        with self.assertNoLogs(logger=logger):
            adapter.gen(n=1)
        # With if we mock the predictions, we should get a log.
        mock_predictions = [
            {m: [10.0] for m in exp.metrics},
            {m: {m: [0.0]} for m in exp.metrics},
        ]
        with mock.patch.object(
            adapter, "predict", return_value=mock_predictions
        ), self.assertLogs(logger=logger) as mock_logs:
            adapter.gen(n=1)
        self.assertIn("deviate more than", mock_logs.records[0].getMessage())



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import pandas as pd
from ax.adapter.factory import get_sobol
from ax.analysis.healthcheck.regression_detection_utils import (
    detect_regressions_by_trial,
    detect_regressions_single_trial,
)
from ax.core.data import Data
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment_with_multi_objective,
    TEST_SOBOL_SEED,
)


class TestRegressionDetection(TestCase):
    def test_regression_detection_by_trial(self) -> None:
        experiment = get_branin_experiment_with_multi_objective(
            with_batch=True, with_status_quo=True
        )

        df0 = pd.DataFrame(
            {
                "metric_name": ["branin_a"] * 6 + ["branin_b"] * 6,
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"] * 2,
                "trial_index": [0] * 12,
                "mean": list(np.arange(7, 1, -1)) + list(np.arange(6, 1, -1)) + [10.0],
                "sem": [1.0] * 12,
            }
        )

        # zero size threshold and high probability threshold
        regressing_arms_metrics = detect_regressions_single_trial(
            experiment=experiment,
            thresholds={"branin_a": (0.0, 0.90), "branin_b": (0.0, 0.90)},
            data=Data(df=df0),
        )
        self.assertGreaterEqual(regressing_arms_metrics["0_4"]["branin_b"], 0.90)
        self.assertEqual(len(list(regressing_arms_metrics.keys())), 1)
        self.assertEqual(len(list(regressing_arms_metrics["0_4"].keys())), 1)
        self.assertGreaterEqual(regressing_arms_metrics["0_4"]["branin_b"], 0.90)

        # zero size threshold and different probability threshold
        regressing_arms_metrics = detect_regressions_single_trial(
            experiment=experiment,
            thresholds={"branin_a": (0, 0.10), "branin_b": (0, 0.95)},
            data=Data(df=df0),
        )
        self.assertGreaterEqual(regressing_arms_metrics["0_0"]["branin_a"], 0.10)
        self.assertEqual(len(list(regressing_arms_metrics.keys())), 1)
        self.assertEqual(len(list(regressing_arms_metrics["0_0"].keys())), 1)

        # different size thresholds
        regressing_arms_metrics = detect_regressions_single_trial(
            experiment=experiment,
            data=Data(df=df0),
            thresholds={"branin_a": (10.0, 0.50), "branin_b": (100, 0.50)},
        )
        self.assertEqual(len(regressing_arms_metrics), 0)

        # add one more trial
        sobol_generator = get_sobol(
            search_space=experiment.search_space, seed=TEST_SOBOL_SEED + 1
        )
        sobol_run = sobol_generator.gen(n=3)
        experiment.new_batch_trial(should_add_status_quo_arm=True).add_generator_run(
            sobol_run
        )

        df1 = pd.DataFrame(
            {
                "metric_name": ["branin_a"] * 4 + ["branin_b"] * 4,
                "arm_name": ["status_quo", "1_0", "1_1", "1_2"] * 2,
                "trial_index": [1] * 8,
                "mean": list(np.arange(5, 1, -1)) + list(np.arange(4, 1, -1)) + [20.0],
                "sem": [1.0] * 8,
            }
        )

        df = pd.concat([df0, df1], ignore_index=True)
        regressing_arms_metrics_by_trial = detect_regressions_by_trial(
            experiment=experiment,
            thresholds={"branin_a": (0.0, 0.90), "branin_b": (0.0, 0.90)},
            data=Data(df=df),
        )
        self.assertGreaterEqual(
            regressing_arms_metrics_by_trial[0]["0_4"]["branin_b"], 0.90
        )
        self.assertEqual(len(list(regressing_arms_metrics_by_trial[0].keys())), 1)
        self.assertEqual(
            len(list(regressing_arms_metrics_by_trial[0]["0_4"].keys())), 1
        )
        self.assertGreaterEqual(
            regressing_arms_metrics_by_trial[1]["1_2"]["branin_b"], 0.90
        )
        self.assertEqual(len(list(regressing_arms_metrics_by_trial[1].keys())), 1)
        self.assertEqual(
            len(list(regressing_arms_metrics_by_trial[1]["1_2"].keys())), 1
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.adapter.base import Adapter

from ax.analysis.analysis import Analysis

from ax.analysis.healthcheck.healthcheck_analysis import HealthcheckAnalysisCard
from ax.core.experiment import Experiment
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.common.testutils import TestCase

ERROR_MESSAGE = "Dummy analysis failed!"


class TestHealtheckException(TestCase):
    class DummyAnalysis(Analysis):
        def compute(
            self,
            experiment: Experiment | None = None,
            generation_strategy: GenerationStrategy | None = None,
            adapter: Adapter | None = None,
        ) -> HealthcheckAnalysisCard:
            raise ValueError(ERROR_MESSAGE)

    def test_error_analysis_card_on_exception(self) -> None:
        analysis = self.DummyAnalysis()
        with self.assertLogs("ax.analysis.analysis", "ERROR") as logs:
            analysis_cards = analysis.compute_or_error_card().flatten()
        # Check that the error message is logged
        self.assertEqual(len(logs.output), 2)
        self.assertIn(ERROR_MESSAGE, logs.output[1])

        # Check that a failed healtheck analysis card is produced
        self.assertEqual(len(analysis_cards), 1)
        analysis_card = analysis_cards[0]

        # Check that error message is in the analysis card
        self.assertEqual(analysis_card.name, "DummyAnalysis")
        self.assertEqual(analysis_card.title, "DummyAnalysis Error")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Sequence

import numpy as np

import pandas as pd
import torch
from ax.adapter.base import Adapter
from ax.adapter.registry import Generators
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.observation import ObservationFeatures
from ax.core.outcome_constraint import OutcomeConstraint, ScalarizedOutcomeConstraint
from ax.core.trial_status import TrialStatus
from ax.core.types import ComparisonOp
from ax.core.utils import get_target_trial_index
from ax.exceptions.core import AxError, UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.stats.statstools import relativize
from botorch.utils.probability.utils import compute_log_prob_feas_from_bounds
from pyre_extensions import none_throws

logger: Logger = get_logger(__name__)

# Warn if p_feasible is less than this threshold.
# TODO: Move this constrant to best point utilities so that the logic for warning in
# analyses is also gates points from being chosen during best point selection.
POSSIBLE_CONSTRAINT_VIOLATION_THRESHOLD: float = 0.05


def extract_relevant_adapter(
    experiment: Experiment | None,
    generation_strategy: GenerationStrategy | None,
    adapter: Adapter | None,
) -> Adapter:
    """
    Analysis.compute(...) takes in both a GenerationStrategy and an Adapter as optional
    arguments. To keep consistency we always want use the same logic for extracting the
    relevant adapter across all Analysis classes.

    * If an Adapter is provided, return it.
    * If a GenerationStrategy is provided, return the Adapter from the current
        GenerationNode. Additionally, if it has not been fit, fit it using the
        Experiment.
    * If both are provided prefer the Adapter.
    * If neither are provided, raise an Exception.
    """

    if generation_strategy is None and adapter is None:
        raise UserInputError(
            "Must provide either a GenerationStrategy or an Adapter to compute this "
            "Analysis."
        )

    if adapter is not None:
        return adapter

    generation_strategy = none_throws(generation_strategy)

    if (adapter := generation_strategy.adapter) is not None:
        return adapter

    if experiment is None:
        raise UserInputError(
            "Provided GenerationStrategy has no adapter, but no Experiment was "
            "provided to source data to fit the adapter."
        )

    generation_strategy.current_node._fit(experiment=experiment)

    return none_throws(generation_strategy.adapter)


def prepare_arm_data(
    experiment: Experiment,
    metric_names: Sequence[str],
    use_model_predictions: bool,
    adapter: Adapter | None = None,
    trial_index: int | None = None,
    trial_statuses: Sequence[TrialStatus] | None = None,
    additional_arms: Sequence[Arm] | None = None,
    relativize: bool = False,
) -> pd.DataFrame:
    """
    Create a table with one entry per arm and columns for each requested metric. This
    is useful for all analyses where the atomic unit is an Arm (ex. ArmEffectsPlot,
    ScatterPlot, ParallelCoordinatesPlot).

    Args:
        experiment: The experiment to extract data from.
        metric_names: The names of the metrics to include in the table.
        use_model_predictions: Whether to use an Adapter to predict the effects of each
            arm. This is often more trustworthy (and leads to better reproducibility)
            than using the raw data, especially when model fit is good and in high-
            noise settings.
        adapter: The adapter to use to predict the effects of each arm if using model
            predictions.
        trial_index: If present, only use arms from the trial with the given index.
            Otherwise include all arms on the experiment. If trial_index=-1, do not
            include any arms from the experiment.
        trial_statuses: If present, only include arms from trials with statuses in this
            collection. If not present, include all arms on the experiment.
        additional_arms: If present, include these arms in the table. These arms will
            be marked as belonging to a trial with index -1.
        relativize: Whether to relativize the effects of each arm against the status
            quo arm. If multiple status quo arms are present, relativize each arm
            against the status quo arm from the same trial.

    Returns a DataFrame with the following columns:
        - trial_index
        - arm_name
        - generation_node
        - METRIC_NAME_mean for each metric_name in metric_names
        - METRIC_NAME_sem for each metric_name in metric_names
    """

    # Ensure a valid combination of arguments is provided.
    if len(metric_names) < 1:
        raise UserInputError("Must provide at least one metric name.")

    missing_metrics = (
        set(metric_names) - set(experiment.metrics.keys()) - {"p_feasible"}
    )
    if missing_metrics:
        raise UserInputError(
            f"Requested metrics {missing_metrics} are not present in the experiment."
        )

    if (
        trial_index is not None
        # -1 is allowed and signifies that no arms from the experiment should be
        # included.
        and trial_index != -1
        and trial_index not in experiment.trials.keys()
    ):
        raise UserInputError(f"Trial with index {trial_index} not found in experiment.")

    if trial_index is not None and trial_statuses is not None:
        raise UserInputError(
            "Cannot provide both trial_index and trial_statuses. Either provide a "
            "trial_index to filter on or a collection of trial_statuses to filter on."
        )

    # Compute the trial index of the target trial both to pass as a fixed feature
    # during prediction if using model predictions, and to relativize against the
    # status quo arm from the target trial if relativizing.
    target_trial_index = get_target_trial_index(experiment=experiment)
    filtered_metric_names = [name for name in metric_names if name != "p_feasible"]
    if use_model_predictions:
        if adapter is None:
            raise UserInputError(
                "Must provide an adapter to use model predictions for the analysis."
            )

        if not adapter.can_model_in_sample:
            logger.info(
                "Provided adapter is unable to model effects. "
                "Using Empirical Bayes as falback."
            )

            trial_indices = None  # This will indicate all trials to `lookup_data`
            if trial_index is not None:
                trial_indices = {trial_index}
                # always look up target trial data even if it's outside the filters
                if target_trial_index is not None:
                    trial_indices.add(target_trial_index)
            data = experiment.lookup_data(trial_indices=trial_indices)
            adapter = Generators.EMPIRICAL_BAYES_THOMPSON(
                experiment=experiment, data=data
            )

        df = _prepare_modeled_arm_data(
            experiment=experiment,
            metric_names=filtered_metric_names,
            adapter=adapter,
            trial_index=trial_index,
            trial_statuses=trial_statuses,
            additional_arms=additional_arms,
            target_trial_index=target_trial_index,
        )
    else:
        if additional_arms is not None:
            raise UserInputError(
                "Cannot provide additional arms when use_model_predictions=False since"
                "there is no observed raw data for the additional arms that are not "
                "part of the Experiment."
            )

        df = _prepare_raw_arm_data(
            metric_names=filtered_metric_names,
            experiment=experiment,
            trial_index=trial_index,
            trial_statuses=trial_statuses,
            target_trial_index=target_trial_index,
        )
    raw_df = df
    has_relative_constraints = experiment.optimization_config is not None and any(
        oc.relative for oc in experiment.optimization_config.outcome_constraints
    )
    status_quo_df = None
    if relativize or has_relative_constraints:
        is_raw_data = not use_model_predictions
        status_quo_df = _get_status_quo_df(
            experiment=experiment,
            df=raw_df,
            metric_names=metric_names,
            is_raw_data=is_raw_data,
            trial_index=trial_index,
            trial_statuses=trial_statuses,
            target_trial_index=target_trial_index,
        )
        if relativize:
            df = relativize_data(
                experiment=experiment,
                df=df,
                metric_names=filtered_metric_names,
                is_raw_data=is_raw_data,
                trial_index=trial_index,
                trial_statuses=trial_statuses,
                target_trial_index=target_trial_index,
                status_quo_df=status_quo_df,
            )

    # Add additional columns which do not require predicting or extracting data.
    df["trial_status"] = df["trial_index"].apply(
        lambda trial_index: experiment.trials[trial_index].status.name
        if trial_index != -1
        else "Additional Arm"
    )

    df["generation_node"] = df.apply(
        lambda row: _extract_generation_node_name(
            trial=experiment.trials[row["trial_index"]],
            arm=experiment.arms_by_name[row["arm_name"]],
        )
        if row["trial_index"] in experiment.trials.keys()
        else Keys.UNKNOWN_GENERATION_NODE.value,
        axis=1,
    )
    df["p_feasible_mean"] = _prepare_p_feasible(
        df=raw_df,
        status_quo_df=status_quo_df,
        outcome_constraints=experiment.optimization_config.outcome_constraints
        if experiment.optimization_config is not None
        else [],
    )
    df["p_feasible_sem"] = np.nan

    # Earlier we add target trial data to the df to support relativization easily
    # but if a specific trial, or trial statuses were requested, and the target trial
    # isn't in that subset, let's remove it so it isn't plotted
    if target_trial_index is not None:
        if trial_index is not None and trial_index != target_trial_index:
            df = df[df["trial_index"] != target_trial_index]
        elif (
            trial_statuses is not None
            and experiment.trials[target_trial_index].status not in trial_statuses
        ):
            df = df[df["trial_index"] != target_trial_index]

    return df


def _prepare_modeled_arm_data(
    experiment: Experiment,
    metric_names: Sequence[str],
    adapter: Adapter,
    trial_index: int | None = None,
    trial_statuses: Sequence[TrialStatus] | None = None,
    additional_arms: Sequence[Arm] | None = None,
    target_trial_index: int | None = None,
) -> pd.DataFrame:
    """
    Compute the modeled (mean, sem) for each arm for each requested metric.

    Returns a DataFrame with the following columns:
        - trial_index
        - arm_name
        - METRIC_NAME_mean for each metric_name in metric_names
        - METRIC_NAME_sem for each metric_name in metric_names
    """
    # Extract the information necessary to construct each row of the DataFrame.

    # Filter trials by trial_index, target_trial_index, and trial_statuses
    filtered_trials = [
        trial
        for trial in experiment.trials.values()
        if (
            (
                (trial.index == trial_index)
                or (trial_index is None)
                or (trial.index == target_trial_index)
            )
            and ((trial_statuses is None) or (trial.status in trial_statuses))
        )
    ]

    # Exclude abandoned arms if the trial is of type BatchTrial
    # https://www.internalfb.com/code/fbsource/[a19525e3f9e6]/fbcode/ax/core/batch_trial.py?lines=51
    # Note: abandoned arms are expected to be excluded in all modeling and predictions.
    # If there is a use case to include abandoned arms, please modify this logic.
    trial_index_arm_pairs = [
        (trial.index, arm)
        for trial in filtered_trials
        for arm in (
            [
                arm
                for arm in trial.arms
                if arm.name
                not in [
                    abandoned_arm.name
                    for abandoned_arm in trial.abandoned_arms_metadata
                ]
            ]
            if isinstance(trial, BatchTrial)
            else trial.arms
        )
    ]
    # Add additional arms passed in by the user
    trial_index_arm_pairs += [(-1, arm) for arm in additional_arms or []]

    # Remove arms with missing parameters since we cannot predict for them.
    predictable_pairs = [
        (trial_index, arm)
        for trial_index, arm in trial_index_arm_pairs
        # Remove arms outside of the search space since we cannot predict for them.
        if experiment.search_space.check_membership(
            parameterization=arm.parameters,
            raise_error=False,
            check_all_parameters_present=True,
        )
    ]
    unpredictable_pairs = [
        (trial_index, arm)
        for trial_index, arm in trial_index_arm_pairs
        if not experiment.search_space.check_membership(
            parameterization=arm.parameters,
            raise_error=False,
            check_all_parameters_present=True,
        )
    ]

    # Batch predict for efficiency.
    predictions = adapter.predict(
        observation_features=[
            ObservationFeatures.from_arm(
                arm=arm,
                # Always predict as if the arm was run in the target trial.
                trial_index=target_trial_index,
            )
            for _, arm in predictable_pairs
        ]
    )
    filtered_metric_names = [name for name in metric_names if name != "p_feasible"]
    records = [
        *[
            {
                "trial_index": predictable_pairs[i][0],
                "arm_name": predictable_pairs[i][1].name
                if predictable_pairs[i][1].has_name
                else f"{Keys.UNNAMED_ARM.value}_{i}",
                **{
                    f"{metric_name}_mean": predictions[0][metric_name][i]
                    for metric_name in filtered_metric_names
                },
                **{
                    f"{metric_name}_sem": predictions[1][metric_name][metric_name][i]
                    ** 0.5
                    for metric_name in filtered_metric_names
                },
            }
            for i in range(len(predictable_pairs))
        ],
        *[
            {
                "trial_index": unpredictable_pairs[i][0],
                "arm_name": unpredictable_pairs[i][1].name
                if unpredictable_pairs[i][1].has_name
                else f"{Keys.UNNAMED_ARM.value}_{i}",
                **{
                    f"{metric_name}_mean": None for metric_name in filtered_metric_names
                },
                **{f"{metric_name}_sem": None for metric_name in filtered_metric_names},
            }
            for i in range(len(unpredictable_pairs))
        ],
    ]

    return pd.DataFrame.from_records(records)


def _prepare_raw_arm_data(
    metric_names: Sequence[str],
    experiment: Experiment,
    trial_index: int | None,
    trial_statuses: Sequence[TrialStatus] | None,
    target_trial_index: int | None,
) -> pd.DataFrame:
    """
    Extract the raw (mean, sem) for each arm for each requested metric.

    Returns a DataFrame with the following columns:
        - trial_index
        - arm_name
        - METRIC_NAME_mean for each metric_name in metric_names
        - METRIC_NAME_sem for each metric_name in metric_names
    """
    # Trial index of -1 indicates candidate arms, we don't have observed data for
    # hypothetical arms so return an empty df
    if trial_index == -1:
        return pd.DataFrame()
    else:
        trials = [
            trial
            for trial in experiment.trials.values()
            if (trial.index == trial_index or trial_index is None)
            and ((trial_statuses is None) or (trial.status in trial_statuses))
            or (trial.index == target_trial_index)
        ]
        data_df = experiment.lookup_data(
            trial_indices=[trial.index for trial in trials]
        ).df

    records = []
    for trial in trials:
        for arm in trial.arms:
            # Extract (mean, sem) pairs for each metric when available, otherwise set to
            # None.
            means = {}
            sems = {}
            for metric_name in metric_names:
                mask = (
                    (data_df["trial_index"] == trial.index)
                    & (data_df["arm_name"] == arm.name)
                    & (data_df["metric_name"] == metric_name)
                )

                if not mask.any():
                    means[metric_name] = None
                    sems[metric_name] = None
                else:
                    means[metric_name] = data_df.loc[mask, "mean"].iloc[0]
                    sems[metric_name] = data_df.loc[mask, "sem"].iloc[0]

            records.append(
                {
                    "trial_index": trial.index,
                    "arm_name": arm.name,
                    **{
                        f"{metric_name}_mean": means[metric_name]
                        for metric_name in metric_names
                    },
                    **{
                        f"{metric_name}_sem": sems[metric_name]
                        for metric_name in metric_names
                    },
                }
            )

    return pd.DataFrame.from_records(records)


def _extract_generation_node_name(trial: BaseTrial, arm: Arm) -> str:
    """
    Extract the name of the GenerationNode that generated the given arm, assuming the
    arm was generated as part of the given trial.

    Args:
        trial: The trial to extract the GenerationNode from.
        arm: The arm to extract the GenerationNode from.

    Returns:
        The name of the GenerationNode that generated the given arm.
    """

    for gr in trial.generator_runs:
        if arm.signature in gr.arm_signatures:
            return gr._generation_node_name or Keys.UNKNOWN_GENERATION_NODE.value

    return Keys.UNKNOWN_GENERATION_NODE.value


def _prepare_p_feasible(
    df: pd.DataFrame,
    status_quo_df: pd.DataFrame | None,
    outcome_constraints: Sequence[OutcomeConstraint],
) -> pd.Series:
    """
    Compute the probability that each arm is feasible with respect to the given
    outcome constraints (assuming normally distributed observations). Calculated in a
    batch for efficiency.


    Args:
        df: Result of _prepare_modeled_arm_data or _prepare_raw_arm_data.
        outcome_constraints: The outcome constraints to use to compute the probability
            of feasibility.

    Returns:
        A Series with one entry per row in df describing the probability that the arm
        is feasible with respect to the given outcome constraints.
    """
    rel_df = None
    if any(c.relative for c in outcome_constraints):
        if status_quo_df is None:
            raise AxError("Must provide status quo data to relativize data.")
        rel_df = _relativize_df_with_sq(
            df=df, status_quo_df=status_quo_df, as_percent=True
        )
    if len(outcome_constraints) == 0:
        return pd.Series(np.ones(len(df)))

    # If an arm is missing data for a metric leave the mean as NaN.
    oc_names = []
    for oc in outcome_constraints:
        if isinstance(oc, ScalarizedOutcomeConstraint):
            # take the str representation of the scalarized outcome constraint
            oc_names.append(str(oc))
        else:
            oc_names.append(oc.metric.name)

    assert len(oc_names) == len(outcome_constraints)

    means = []
    sigmas = []
    for i, oc_name in enumerate(oc_names):
        df_constraint = none_throws(rel_df if outcome_constraints[i].relative else df)
        # TODO[T235432214]: currently we are leaving the mean as NaN if the constraint
        # is on ScalarizedOutcomeConstraint but we should be able to calculate it by
        # setting the mean to be weights * individual metrics and sem to be
        # sqrt(sum((weights * individual_sems)^2)), assuming independence.
        if f"{oc_name}_mean" in df_constraint.columns:
            means.append(df_constraint[f"{oc_name}_mean"].tolist())

        else:
            means.append(np.nan * np.ones(len(df_constraint)))
        sigmas.append(
            (df_constraint[f"{oc_name}_sem"].fillna(0)).tolist()
            if f"{oc_name}_sem" in df_constraint.columns
            else [0] * len(df)
        )

    con_lower_inds = [
        i
        for i in range(len(outcome_constraints))
        if outcome_constraints[i].op == ComparisonOp.GEQ
    ]
    con_upper_inds = [
        i
        for i in range(len(outcome_constraints))
        if outcome_constraints[i].op == ComparisonOp.LEQ
    ]

    con_lower = [
        constraint.bound
        for constraint in outcome_constraints
        if constraint.op == ComparisonOp.GEQ
    ]

    con_upper = [
        constraint.bound
        for constraint in outcome_constraints
        if constraint.op == ComparisonOp.LEQ
    ]

    log_prob_feas = compute_log_prob_feas_from_bounds(
        con_lower_inds=torch.tensor(con_lower_inds, dtype=torch.int),
        con_upper_inds=torch.tensor(con_upper_inds, dtype=torch.int),
        con_lower=torch.tensor(con_lower, dtype=torch.double),
        con_upper=torch.tensor(con_upper, dtype=torch.double),
        con_both_inds=torch.empty(0, dtype=torch.int),
        con_both=torch.empty(0),
        means=torch.tensor(means, dtype=torch.double).T,
        sigmas=torch.tensor(sigmas, dtype=torch.double).T,
    )

    return pd.Series(log_prob_feas.exp())


def _relativize_df_with_sq(
    df: pd.DataFrame,
    status_quo_df: pd.DataFrame,
    as_percent: bool = False,
) -> pd.DataFrame:
    """
    Relativize the data with respect to some status quo arm.

    Args:
        df: DataFrame containing the data to be relativized. Must include columns
            'METRIC_NAME_mean' and 'METRIC_NAME_sem' for each metric, as well as a
            column identifying the trial and a column identifying the status quo
            arm for each trial.
        status_quo_df: DataFrame containing the status quo data for each trial.

    Returns:
        A DataFrame with the same structure as the input, but with 'METRIC_NAME_mean'
        and 'METRIC_NAME_sem' columns relativized to the status quo arm for each metric
        within each trial.
    """
    metric_names = [name[:-5] for name in df.columns if name.endswith("_mean")]

    rel_df = df.copy()

    for trial_idx, trial_df in rel_df.groupby("trial_index"):
        status_quo_row = status_quo_df[status_quo_df["trial_index"] == trial_idx]

        for metric_name in metric_names:
            mean_col = f"{metric_name}_mean"
            sem_col = f"{metric_name}_sem"

            y_rel, y_se_rel = relativize(
                means_t=trial_df[mean_col],
                sems_t=trial_df[sem_col],
                mean_c=status_quo_row[mean_col].values[0],
                sem_c=status_quo_row[sem_col].values[0],
                as_percent=as_percent,
            )

            rel_df.loc[rel_df["trial_index"] == trial_idx, mean_col] = y_rel
            rel_df.loc[rel_df["trial_index"] == trial_idx, sem_col] = y_se_rel

    return rel_df


def _get_sq_arm_name(experiment: Experiment) -> str:
    """
    Retrieve the name of the status quo arm from the given experiment.

    Args:
        experiment: An Ax experiment.
    Returns:
        The name of the status quo arm.
    Raises:
        UserInputError: If the experiment does not have a status quo arm or if the
        status quo arm is not named.

    """
    if experiment.status_quo is None:
        raise UserInputError(
            "Cannot relativize data without a status quo arm on the experiment."
        )

    if experiment.status_quo.name is None:
        raise UserInputError(
            "Cannot relativize data without a named status quo arm on the "
            "experiment."
        )

    return experiment.status_quo.name


def relativize_data(
    experiment: Experiment,
    df: pd.DataFrame,
    metric_names: Sequence[str],
    is_raw_data: bool,
    trial_index: int | None,
    trial_statuses: Sequence[TrialStatus] | None,
    target_trial_index: int | None,
    status_quo_df: pd.DataFrame | None = None,
) -> pd.DataFrame:
    """
    Relativize the data in the given DataFrame with respect to the status quo arm.
    This method includes logic to select the appropriate status quo arm to use
    for relativization. If relativizing raw data, the status quo is taken from
    the inputted trial and falls back to the only status quo arm on the experiment
    if the former is not avaialble. If relativizing modeled data, the status quo is
    taken from the target trial and falls back to the raw status quo arm on the
    target trial if the former is not available.

    Args:
        experiment: An Ax experiment.
        df: The DataFrame containing the data to be relativized. Must include columns
            'METRIC_NAME_mean', 'METRIC_NAME_sem', trial_index, and arm_name.
        metric_names: The names of the metrics to relativize.
        is_raw_data: Whether the data is raw or modeled.
        trial_index: If present, only relativize data from the trial with the given
            index.
        trial_statuses: If present, only relativize data from trials with statuses in
            this collection.
        target_trial_index: The index of the target trial.
        status_quo_df: The status quo data for each trial.
    Returns:
        A new DataFrame with the same structure as the input, but with
            'METRIC_NAME_mean'
        and 'METRIC_NAME_sem' columns relativized to the status quo arm for
            each metric
        within each trial.
    Raises:
        UserInputError: If the experiment does not have a status quo arm, if the status
            quo arm is not named, or if no status quo arm is found in the experiment.
    """
    if status_quo_df is None:
        status_quo_df = _get_status_quo_df(
            experiment=experiment,
            df=df,
            metric_names=metric_names,
            is_raw_data=is_raw_data,
            trial_index=trial_index,
            trial_statuses=trial_statuses,
            target_trial_index=target_trial_index,
        )
    return _relativize_df_with_sq(df=df, status_quo_df=status_quo_df)


def _get_status_quo_df(
    experiment: Experiment,
    df: pd.DataFrame,
    metric_names: Sequence[str],
    is_raw_data: bool,
    trial_index: int | None,
    trial_statuses: Sequence[TrialStatus] | None,
    target_trial_index: int | None,
) -> pd.DataFrame:
    status_quo_name = _get_sq_arm_name(experiment=experiment)

    if not is_raw_data:
        # Use the status quo arm from the target trial, prefer model predictions
        # over raw observations.
        mask = (df["arm_name"] == status_quo_name) & (
            df["trial_index"] == target_trial_index
        )

        # Use the raw status quo effects if no model predictions are available
        # (ex. if the status quo had None in its parameters).
        if not mask.any() or df[mask].isna().any().any():
            raw_df = _prepare_raw_arm_data(
                metric_names=metric_names,
                experiment=experiment,
                trial_index=trial_index,
                trial_statuses=trial_statuses,
                target_trial_index=target_trial_index,
            )

            # Fallback on the raw observations from the status quo arm on the
            # target trial.
            raw_mask = (raw_df["arm_name"] == status_quo_name) & (
                raw_df["trial_index"] == target_trial_index
            )

            if not raw_mask.any():
                raise UserInputError(
                    "Could not find a status quo arm on Trial "
                    f"{target_trial_index} to relativize against."
                )

            status_quo_row = raw_df[raw_mask]
        else:
            status_quo_row = df[mask]

        status_quo_df = pd.concat([status_quo_row] * len(df["trial_index"].unique()))
        status_quo_df["trial_index"] = df["trial_index"].unique()

    else:
        # If not using model predictions search for an appropriate status quo arm
        # using the following logic:
        # 1. Use the status quo arm from the same trial as the arm being
        #    relativized.
        # 2. Use the only status quo arm on the experiment.
        # 3. Raise an exception
        status_quo_mask = df["arm_name"] == status_quo_name
        status_quo_rows = []
        for trial_idx in df["trial_index"].unique():
            trial_mask = (df["trial_index"] == trial_idx) & status_quo_mask
            if trial_mask.any():
                row = df[trial_mask].iloc[0]
            elif status_quo_mask.sum() == 1:
                row = df[status_quo_mask].iloc[0]
            else:
                raise UserInputError(
                    "Failed to relativize, no status quo arm found in the "
                    "experiment."
                )

            row["trial_index"] = trial_idx
            status_quo_rows.append(row)

        status_quo_df = pd.DataFrame(status_quo_rows)[
            [
                "trial_index",
                *[f"{name}_mean" for name in metric_names if name != "p_feasible"],
                *[f"{name}_sem" for name in metric_names if name != "p_feasible"],
            ]
        ]
    return status_quo_df


def get_lower_is_better(experiment: Experiment, metric_name: str) -> bool | None:
    if metric_name == "p_feasible":
        return False
    return experiment.metrics[metric_name].lower_is_better


def update_metric_names_if_using_p_feasible(
    metric_names: Sequence[str], experiment: Experiment
) -> list[str]:
    """Return a new list of unique metric names including metrics in
    metric_names and metrics involved in constraints if using p(feasible).
    """
    unique_metric_names = set(metric_names)
    if "p_feasible" in unique_metric_names:
        if "p_feasible" in experiment.metrics.keys():
            raise AxError(
                "p_feasible is reserved for plotting the probability of"
                " feasibility of an arm with respect to outcome constraints"
                " in Ax AnalysisCards, but there is a name collision with a "
                " metric named p_feasible on the experiment."
            )
        optimization_config = none_throws(
            experiment.optimization_config,
            "Cannot compute p_feasible without an optimization config.",
        )
        for oc in optimization_config.outcome_constraints:
            unique_metric_names.add(oc.metric.name)
    return list(unique_metric_names)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from itertools import product
from unittest.mock import Mock

import numpy as np
import pandas as pd
from ax.benchmark.benchmark_metric import (
    _get_no_metadata_msg,
    BenchmarkMapMetric,
    BenchmarkMapUnavailableWhileRunningMetric,
    BenchmarkMetric,
    BenchmarkMetricBase,
    BenchmarkTimeVaryingMetric,
)
from ax.benchmark.benchmark_trial_metadata import BenchmarkTrialMetadata
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.map_metric import MapMetric
from ax.core.metric import MetricFetchE
from ax.core.trial import Trial
from ax.utils.common.result import Err
from ax.utils.common.testutils import TestCase
from ax.utils.testing.backend_simulator import BackendSimulator, BackendSimulatorOptions
from ax.utils.testing.core_stubs import get_experiment
from pyre_extensions import assert_is_instance, none_throws


def _get_one_step_df(
    batch: bool, metric_name: str, step: int, observe_noise_sd: bool
) -> pd.DataFrame:
    if observe_noise_sd:
        sem = [0.1, 0.0] if batch else [0.1]
    else:
        sem = [np.nan, np.nan] if batch else [np.nan]
    if metric_name == "test_metric1":
        return pd.DataFrame(
            {
                "arm_name": ["0_0", "0_1"] if batch else ["0_0"],
                "metric_name": metric_name,
                "mean": [1.0, 2.5] if batch else [1.0],
                "sem": sem,
                "trial_index": 0,
                "step": step,
                "virtual runtime": step,
            }
        )
    return pd.DataFrame(
        {
            "arm_name": ["0_0", "0_1"] if batch else ["0_0"],
            "metric_name": metric_name,
            "mean": [0.5, 1.5] if batch else [0.5],
            "sem": sem,
            "trial_index": 0,
            "step": step,
            "virtual runtime": step,
        }
    )


def get_test_trial(
    multiple_time_steps: bool = False,
    batch: bool = False,
    has_metadata: bool = True,
    has_simulator: bool = False,
) -> Trial | BatchTrial:
    experiment = get_experiment()

    arm1 = Arm(parameters={"w": 1.0, "x": 1, "y": "foo", "z": True}, name="0_0")
    arm2 = Arm(parameters={"w": 1.0, "x": 2, "y": "foo", "z": True}, name="0_1")

    if batch:
        trial = experiment.new_batch_trial()
        trial.add_arms_and_weights(arms=[arm1, arm2])
    else:
        trial = experiment.new_trial()
        trial.add_arm(arm=arm1)

    if not has_metadata:
        return trial

    n_steps = 3 if multiple_time_steps else 1
    dfs = {
        name: pd.concat(
            _get_one_step_df(
                batch=batch, metric_name=name, step=i, observe_noise_sd=True
            )
            for i in range(n_steps)
        )
        for name in ["test_metric1", "test_metric2"]
    }

    if has_simulator:
        backend_simulator = BackendSimulator(
            options=BackendSimulatorOptions(
                max_concurrency=1, internal_clock=0, use_update_as_start_time=True
            ),
            verbose_logging=False,
        )
        backend_simulator.run_trial(trial_index=trial.index, runtime=1)
    else:
        backend_simulator = None
    metadata = BenchmarkTrialMetadata(
        dfs=dfs,
        backend_simulator=backend_simulator,
    )

    trial.update_run_metadata({"benchmark_metadata": metadata})
    return trial


class TestBenchmarkMetric(TestCase):
    def setUp(self) -> None:
        self.outcome_names = ["test_metric1", "test_metric2"]
        self.metrics = {
            name: BenchmarkMetric(
                name=name, lower_is_better=True, observe_noise_sd=observe_noise_sd
            )
            for name, observe_noise_sd in zip(
                self.outcome_names, (False, True), strict=True
            )
        }
        self.tv_metrics = {
            name: BenchmarkTimeVaryingMetric(name=name, lower_is_better=True)
            for name in self.outcome_names
        }
        self.map_metrics = {
            name: BenchmarkMapMetric(name=name, lower_is_better=True)
            for name in self.outcome_names
        }
        self.map_unavailable_while_running_metrics = {
            name: BenchmarkMapUnavailableWhileRunningMetric(
                name=name, lower_is_better=True
            )
            for name in self.outcome_names
        }

    def test_available_while_running(self) -> None:
        self.assertFalse(self.metrics["test_metric1"].is_available_while_running())
        self.assertFalse(BenchmarkMetric.is_available_while_running())

        self.assertTrue(self.map_metrics["test_metric1"].is_available_while_running())
        self.assertTrue(BenchmarkMapMetric.is_available_while_running())

        self.assertTrue(self.tv_metrics["test_metric1"].is_available_while_running())
        self.assertTrue(BenchmarkTimeVaryingMetric.is_available_while_running())

        self.assertFalse(
            self.map_unavailable_while_running_metrics[
                "test_metric1"
            ].is_available_while_running()
        )
        self.assertFalse(
            BenchmarkMapUnavailableWhileRunningMetric.is_available_while_running()
        )

    def test_exceptions(self) -> None:
        for metric in [
            self.metrics["test_metric1"],
            self.map_metrics["test_metric1"],
            self.tv_metrics["test_metric1"],
            self.map_unavailable_while_running_metrics["test_metric1"],
        ]:
            with self.subTest(
                f"No-metadata error, metric class={metric.__class__.__name__}"
            ):
                trial = get_test_trial(has_metadata=False)
                result = metric.fetch_trial_data(trial=trial)
                self.assertIsInstance(result, Err)
                self.assertIsInstance(result.value, MetricFetchE)
                self.assertEqual(
                    result.value.message,
                    _get_no_metadata_msg(trial_index=trial.index),
                    msg=f"{metric.__class__.__name__}",
                )

            trial = get_test_trial()
            with self.subTest(
                f"Unsupported kwargs, metric class={metric.__class__.__name__}"
            ), self.assertRaisesRegex(
                NotImplementedError,
                "Arguments {'foo'} are not supported in Benchmark",
            ):
                metric.fetch_trial_data(trial, foo="bar")

        with self.subTest("Error for multiple metrics in BenchmarkMetric"):
            trial = get_test_trial()
            map_trial = get_test_trial(multiple_time_steps=True)
            trial.run_metadata["benchmark_metadata"] = map_trial.run_metadata[
                "benchmark_metadata"
            ]
            with self.assertRaisesRegex(
                ValueError, "Trial has data from multiple time steps"
            ):
                self.metrics["test_metric1"].fetch_trial_data(trial=trial)

    def _test_fetch_trial_data_one_time_step(
        self, batch: bool, metric: BenchmarkMetricBase
    ) -> None:
        trial = get_test_trial(batch=batch, has_simulator=True)
        df1 = assert_is_instance(metric.fetch_trial_data(trial=trial).value, Data).df
        self.assertEqual(len(df1), len(trial.arms))
        expected_results = _get_one_step_df(
            batch=batch,
            metric_name=metric.name,
            step=0,
            observe_noise_sd=metric.observe_noise_sd,
        ).drop(columns=["virtual runtime"])
        if not isinstance(metric, MapMetric):
            expected_results = expected_results.drop(columns=["step"])
        df1_dict = df1.to_dict()
        expected_dict = expected_results.to_dict()
        # Compare SEM separately since NaN equality fails.
        sem, expected_sem = df1_dict.pop("sem"), expected_dict.pop("sem")
        try:
            self.assertEqual(sem, expected_sem)
        except AssertionError:
            self.assertTrue(all(math.isnan(s) for s in sem.values()))
        self.assertDictEqual(df1_dict, expected_dict)

    def test_fetch_trial_data_one_time_step(self) -> None:
        for batch, metrics in product(
            [False, True],
            [
                self.metrics,
                self.map_metrics,
                self.tv_metrics,
                self.map_unavailable_while_running_metrics,
            ],
        ):
            metric = metrics["test_metric1"]
            with self.subTest(
                batch=batch,
                metric_cls=type(metric),
            ):
                self._test_fetch_trial_data_one_time_step(batch=batch, metric=metric)

    def _test_fetch_trial_multiple_time_steps_with_simulator(self, batch: bool) -> None:
        """
        Cases for fetching data with multiple time steps:
        - Metric is 'BenchmarkMetric' -> exception, tested below
        - Has simulator, metric is 'BenchmarkMapMetric' -> df grows with each step
        - No simulator, metric is 'BenchmarkMapMetric' -> all data present while
          running (but realistically it would never be RUNNING)
        - Has simulator, metric is 'BenchmarkTimeVaryingMetric' -> one step at
          a time, evolving as we take steps
        - No simulator, metric is 'BenchmarkTimeVaryingMetric' -> completes
            immediately and returns last step

        See table in benchmark_metric.py for more details.
        """
        metric_name = "test_metric1"

        for metric, has_simulator in product(
            [
                self.map_metrics[metric_name],
                self.tv_metrics[metric_name],
                self.map_unavailable_while_running_metrics[metric_name],
            ],
            [False, True],
        ):
            trial = get_test_trial(
                has_metadata=True,
                batch=batch,
                multiple_time_steps=True,
                has_simulator=has_simulator,
            )
            data = metric.fetch_trial_data(trial=trial).value
            df_or_map_df = data.map_df if isinstance(metric, MapMetric) else data.df
            returns_full_data = (not has_simulator) and isinstance(metric, MapMetric)
            self.assertEqual(
                len(df_or_map_df), len(trial.arms) * (3 if returns_full_data else 1)
            )
            drop_cols = ["virtual runtime"]
            if not isinstance(metric, MapMetric):
                drop_cols += ["step"]

            expected_df = _get_one_step_df(
                batch=batch,
                metric_name=metric_name,
                step=0,
                observe_noise_sd=metric.observe_noise_sd,
            ).drop(columns=drop_cols)
            if returns_full_data:
                self.assertEqual(
                    df_or_map_df[df_or_map_df["step"] == 0].to_dict(),
                    expected_df.to_dict(),
                )
            else:
                self.assertEqual(df_or_map_df.to_dict(), expected_df.to_dict())

            backend_simulator = trial.run_metadata[
                "benchmark_metadata"
            ].backend_simulator
            self.assertEqual(backend_simulator is None, not has_simulator)
            if backend_simulator is None:
                continue
            self.assertEqual(backend_simulator.time, 0)
            sim_trial = none_throws(
                backend_simulator.get_sim_trial_by_index(trial.index)
            )
            self.assertIn(sim_trial, backend_simulator._running)
            backend_simulator.update()
            self.assertEqual(backend_simulator.time, 1)
            self.assertIn(sim_trial, backend_simulator._completed)
            backend_simulator.update()
            self.assertIn(sim_trial, backend_simulator._completed)
            self.assertEqual(backend_simulator.time, 2)
            data = metric.fetch_trial_data(trial=trial).value
            if isinstance(metric, MapMetric):
                map_df = data.map_df
                self.assertEqual(len(map_df), 2 * len(trial.arms))
                self.assertEqual(set(map_df["step"].tolist()), {0, 1})
            df = data.df
            self.assertEqual(len(df), len(trial.arms))
            expected_df = _get_one_step_df(
                batch=batch,
                metric_name=metric_name,
                step=1,
                observe_noise_sd=metric.observe_noise_sd,
            ).drop(columns=drop_cols)
            self.assertEqual(df.reset_index(drop=True).to_dict(), expected_df.to_dict())

    def test_fetch_trial_multiple_time_steps_with_simulator(self) -> None:
        self._test_fetch_trial_multiple_time_steps_with_simulator(batch=False)
        self._test_fetch_trial_multiple_time_steps_with_simulator(batch=True)

    def test_sim_trial_completes_in_future_raises(self) -> None:
        simulator = BackendSimulator()
        simulator.run_trial(trial_index=0, runtime=0)
        simulator.update()
        simulator.options.internal_clock = -1
        metadata = BenchmarkTrialMetadata(
            dfs={"test_metric": pd.DataFrame({"t": [3], "step": 0})},
            backend_simulator=simulator,
        )
        trial = Mock(spec=Trial)
        trial.index = 0
        trial.run_metadata = {"benchmark_metadata": metadata}
        metric = BenchmarkMapMetric(name="test_metric", lower_is_better=True)
        with self.assertRaisesRegex(RuntimeError, "in the future"):
            metric.fetch_trial_data(trial=trial)

    def test_multiple_time_steps_with_BenchmarkMetric_raises(self) -> None:
        trial = get_test_trial(multiple_time_steps=True)
        metric = self.metrics["test_metric1"]
        with self.assertRaisesRegex(ValueError, "data from multiple time steps"):
            metric.fetch_trial_data(trial=trial)

    def test_abandoned_arms_not_supported(self) -> None:
        trial = get_test_trial(batch=True)
        trial.mark_arm_abandoned("0_0")
        with self.assertRaisesRegex(
            NotImplementedError, "does not support abandoned arms"
        ):
            self.map_metrics["test_metric1"].fetch_trial_data(trial=trial)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import random
from typing import Any

import numpy as np
import pandas as pd
import torch
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.exceptions.core import UnsupportedError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import none_throws
from torch import Tensor


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def generic_equals(first: Any, second: Any) -> bool:
    if isinstance(first, Tensor):
        return isinstance(second, Tensor) and torch.equal(first, second)
    if isinstance(first, np.ndarray):
        return isinstance(second, np.ndarray) and np.array_equal(
            first, second, equal_nan=True
        )
    if isinstance(first, dict):
        return isinstance(second, dict) and generic_equals(
            sorted(first.items()), sorted(second.items())
        )
    if isinstance(first, (tuple, list)):
        if type(first) is not type(second) or len(first) != len(second):
            return False
        for f, s in zip(first, second):
            if not generic_equals(f, s):
                return False
        return True
    return first == second


def run_trials_with_gs(
    experiment: Experiment, gs: GenerationStrategy, num_trials: int
) -> None:
    r"""Runs and completes `num_trials` trials for the given experiment with the
    given GS. The trials are completed with random metric values between 0 and 1.

    Args:
        experiment: The experiment to run trials on. Must have an optimization config.
        gs: The generation strategy to use.
        num_trials: The number of trials to run.
    """
    if experiment.optimization_config is None:
        raise UnsupportedError(  # pragma: no cover
            "`run_trials_with_gs` requires the experiment to have "
            "an optimization config."
        )
    existing_trials = len(experiment.trials)
    for i in range(existing_trials, existing_trials + num_trials):
        trial = experiment.new_trial(
            generator_run=gs.gen_single_trial(experiment=experiment)
        )
        data = Data(
            df=pd.DataFrame.from_records(
                [
                    {
                        "arm_name": arm.name,
                        "metric_name": m,
                        "mean": random.random(),
                        "sem": None,
                        "trial_index": i,
                    }
                    for m in none_throws(experiment.optimization_config).metrics
                    for arm in trial.arms
                ]
            )
        )
        experiment.attach_data(data)
        trial.run().complete()
